# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from abc import ABCMeta, abstractmethod, abstractproperty
from logging import getLogger
from os import getenv
from os.path import abspath, isfile
from re import split as resplit
from threading import Lock
from time import sleep, time


__all__ = ("Target", "sanitizer_opts")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]

LOG = getLogger("grizzly")


def sanitizer_opts(env_data):
    """Parse the values defined in given *SAN_OPTIONS environment variable.
    For example "ASAN_OPTIONS=debug=false:log_path='/test/file.log'"
    would return {"debug": "false", "log_path": "'/test/file.log'"}

    Args:
        env_var (str): *SAN_OPTIONS environment variable to parse.

    Returns:
        dict: Sanitized values from environment.
    """
    opts = dict()
    for opt in resplit(r":(?![\\|/])", env_data):
        if not opt:
            continue
        key, val = opt.split("=")
        opts[key] = val
    return opts


class TargetError(Exception):
    """Raised by Target"""


class TargetLaunchError(TargetError):
    """Raised if a failure during launch occurs"""


class TargetLaunchTimeout(TargetError):
    """Raised if the target does not launch within the defined amount of time"""


class Target(metaclass=ABCMeta):
    RESULT_NONE = 0
    RESULT_FAILURE = 1
    RESULT_IGNORED = 2

    __slots__ = (
        "_lock", "_monitor", "binary", "extension", "forced_close", "launch_timeout",
        "log_limit", "memory_limit", "prefs", "rl_countdown", "rl_reset")

    def __init__(self, binary, extension, launch_timeout, log_limit, memory_limit, prefs, relaunch):
        assert log_limit >= 0
        assert memory_limit >= 0
        assert relaunch >= 1
        self._lock = Lock()
        self._monitor = None
        self.binary = binary
        self.extension = extension
        self.forced_close = getenv("GRZ_FORCED_CLOSE") != "0"
        self.launch_timeout = max(launch_timeout, 300)
        self.log_limit = log_limit
        self.memory_limit = memory_limit
        self.prefs = abspath(prefs) if prefs else None
        self.rl_countdown = 0
        self.rl_reset = relaunch
        assert self.binary is not None and isfile(self.binary)
        if self.prefs is not None:
            if not isfile(self.prefs):
                raise TargetError("Prefs file does not exist %r" % (self.prefs,))
            LOG.info("Using prefs %r", self.prefs)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    def add_abort_token(self, token):  # pylint: disable=no-self-use,unused-argument
        LOG.warning("add_abort_token() not implemented!")

    def check_relaunch(self, wait=60):
        if self.rl_countdown > 0:
            return
        # if the adapter does not use the default harness
        # or close the browser it will hang here for 60 seconds
        LOG.debug("relaunch will be triggered... waiting up to %0.2f seconds", wait)
        deadline = time() + wait
        while self.monitor.is_healthy():
            if time() >= deadline:
                LOG.info("Forcing target relaunch")
                break
            sleep(1)
        self.close()

    @abstractmethod
    def cleanup(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractproperty
    def closed(self):
        pass

    @abstractmethod
    def detect_failure(self, ignored, was_timeout):
        pass

    def dump_coverage(self):  # pylint: disable=no-self-use
        LOG.warning("dump_coverage() is not supported!")

    @property
    def expect_close(self):
        # This is used to indicate if the browser will self close after the current iteration
        return self.rl_countdown < 1 and not self.forced_close

    def is_idle(self, threshold):  # pylint: disable=no-self-use,unused-argument
        LOG.debug("Target.is_idle() not implemented! returning False")
        return False

    @abstractmethod
    def launch(self):
        pass

    def log_size(self):  # pylint: disable=no-self-use
        LOG.debug("log_size() not implemented! returning 0")
        return 0

    @abstractproperty
    def monitor(self):
        pass

    def reverse(self, remote, local):
        # remote->device, local->desktop
        pass

    @abstractmethod
    def save_logs(self, *args, **kwargs):
        pass

    def step(self):
        # this should be called once per iteration
        self.rl_countdown -= 1



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from sys import exit as sysexit

from .args import ReplayArgs
from .replay import ReplayManager


sysexit(ReplayManager.main(ReplayArgs().parse_args()))



# coding=utf-8
"""
Sapphire unit tests
"""
# pylint: disable=protected-access

import socket
import threading

import pytest

from .sapphire_job import SapphireJob
from .sapphire_worker import SapphireWorker, SapphireWorkerError

def test_sapphire_worker_01(mocker):
    """test simple SapphireWorker in running state"""
    wthread = mocker.Mock(spec=threading.Thread)
    wthread.is_alive.return_value = True
    worker = SapphireWorker(mocker.Mock(spec=socket.socket), wthread)
    assert worker._conn is not None
    assert worker._thread is not None
    # it is assumed that launch() has already been called at this point
    assert not worker.done
    assert wthread.join.call_count == 0
    assert wthread.is_alive.call_count == 1
    worker.join(timeout=0)
    assert wthread.join.call_count == 1
    assert wthread.is_alive.call_count == 2
    assert worker._conn.close.call_count == 0
    wthread.is_alive.return_value = False
    worker.close()
    assert worker._conn.close.call_count == 1
    assert worker._thread is None
    assert worker.done

def test_sapphire_worker_02(mocker):
    """test simple SapphireWorker fails to close"""
    worker = SapphireWorker(
        mocker.Mock(spec=socket.socket),
        mocker.Mock(spec=threading.Thread))
    # it is assumed that launch() has already been called at this point
    worker._thread.is_alive.return_value = True
    with pytest.raises(SapphireWorkerError, match="Worker thread failed to join!"):
        worker.close()

def test_sapphire_worker_03(mocker):
    """test SapphireWorker.launch() fail cases"""
    serv_con = mocker.Mock(spec=socket.socket)
    serv_job = mocker.Mock(spec=SapphireJob)
    fake_thread = mocker.patch("sapphire.sapphire_worker.threading.Thread", autospec=True)
    mocker.patch("sapphire.sapphire_worker.time.sleep", autospec=True)

    serv_con.accept.side_effect = socket.timeout
    assert SapphireWorker.launch(serv_con, serv_job) is None

    serv_con.accept.side_effect = None
    conn = mocker.Mock(spec=socket.socket)
    serv_con.accept.return_value = (conn, None)
    fake_thread.side_effect = threading.ThreadError
    assert SapphireWorker.launch(serv_con, serv_job) is None
    assert conn.close.call_count == 1
    assert serv_job.accepting.clear.call_count == 0
    assert serv_job.accepting.set.call_count == 1

def test_sapphire_worker_04(mocker, tmp_path):
    """test SapphireWorker.launch()"""
    (tmp_path / "testfile").touch()
    job = SapphireJob(str(tmp_path))
    clnt_sock = mocker.Mock(spec=socket.socket)
    clnt_sock.recv.return_value = b"GET /testfile HTTP/1.1"
    serv_sock = mocker.Mock(spec=socket.socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    worker = SapphireWorker.launch(serv_sock, job)
    assert worker is not None
    try:
        assert job.is_complete(wait=1)
    finally:
        worker.close()
    assert worker.done
    assert serv_sock.accept.call_count == 1
    assert clnt_sock.close.call_count == 2

def test_sapphire_worker_05(mocker):
    """test SapphireWorker.handle_request() socket errors"""
    serv_con = mocker.Mock(spec=socket.socket)
    serv_con.recv.side_effect = socket.error
    serv_job = mocker.Mock(spec=SapphireJob)
    SapphireWorker.handle_request(serv_con, serv_job)
    assert serv_job.accepting.set.call_count == 1
    assert serv_con.sendall.call_count == 0
    assert serv_con.close.call_count == 1

def test_response_data_01():
    """test _200_header()"""
    output = SapphireWorker._200_header(10, "text/html")
    assert b"Content-Length: 10" in output
    assert b"Content-Type: text/html" in output

def test_response_data_02():
    """test _307_redirect()"""
    output = SapphireWorker._307_redirect("http://some.test.url")
    assert b"Location: http://some.test.url" in output

def test_response_data_03():
    """test _4xx_page() without close timeout"""
    output = SapphireWorker._4xx_page(400, "Bad Request")
    assert b"Content-Length: " in output
    assert b"HTTP/1.1 400 Bad Request" in output
    assert b"400!" in output

def test_response_data_04():
    """test _4xx_page() with close timeout"""
    output = SapphireWorker._4xx_page(404, "Not Found", close=10)
    assert b"Content-Length: " in output
    assert b"HTTP/1.1 404 Not Found" in output
    assert b"<script>window.setTimeout(window.close, 10000)</script>" in output



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from os import makedirs
from os.path import join as pathjoin
from tempfile import gettempdir


__all__ = ("grz_tmp",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


def grz_tmp(*subdir):
    path = pathjoin(gettempdir(), "grizzly", *subdir)
    makedirs(path, exist_ok=True)
    return path



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger
from time import sleep, time

from sapphire import SERVED_TIMEOUT
from ..target import TargetLaunchTimeout
from .utils import grz_tmp

__all__ = ("Runner",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger("grz_runner")

# _IdleChecker is used to help determine if the target is hung (actively using CPU)
# or if it has not made expected the HTTP requests for other reasons (idle).
# This will allow the framework to move on without interrupting execution of
# long running test cases.
# This is not perfect! It is to be used AFTER the test case timeout
# (initial_delay) has elapsed.
class _IdleChecker(object):
    __slots__ = ("_check_cb", "_init_delay", "_poll_delay", "_threshold", "_next_poll")

    def __init__(self, check_cb, threshold, initial_delay, poll_delay=1):
        assert callable(check_cb)
        assert initial_delay >= 0
        assert poll_delay >= 0
        assert threshold >= 0
        self._check_cb = check_cb  # callback used to check if target is idle
        self._init_delay = initial_delay  # time to wait before the initial idle poll
        self._poll_delay = poll_delay  # time to wait between subsequent polls
        self._threshold = threshold  # CPU usage threshold
        self._next_poll = None

    def is_idle(self):
        """Check the target idle callback. This is throttled by '_next_poll'
        which specifies the time at which the next callback call is allowed.

        Args:
            None

        Returns:
            bool: True if the target idle callback returned True otherwise False
        """
        assert self._next_poll is not None, "schedule_poll() must be called first"
        now = time()
        if now >= self._next_poll:
            if self._check_cb(self._threshold):
                return True
            self.schedule_poll(now=now)
        return False

    def schedule_poll(self, initial=False, now=None):
        """Update `_next_poll`.

        Args:
            initial (int): Use `_init_delay` to schedule next poll
            now (int): time in seconds

        Returns:
            None
        """
        if now is None:
            now = time()
        if initial:
            self._next_poll = now + self._init_delay
        else:
            self._next_poll = now + self._poll_delay


class Runner(object):
    COMPLETE = 1
    ERROR = 2
    FAILED = 3
    IGNORED = 4

    __slots__ = ("_idle", "_server", "_target", "result", "served", "timeout")

    def __init__(self, server, target, idle_threshold=0, idle_delay=60):
        if idle_threshold > 0:
            self._idle = _IdleChecker(target.is_idle, idle_threshold, idle_delay)
        else:
            self._idle = None
        self._server = server  # a sapphire instance to serve the test case
        self._target = target  # target to run test case
        self.result = None
        self.served = None
        self.timeout = False

    def launch(self, location, env_mod=None, max_retries=3, retry_delay=0):
        """Launch a target and open `location`.

        Args:
            location (str): URL to open via Target.
            env_mod (dict): Environment modifications.
            max_retries (int): Number of retries to preform before re-raising TargetLaunchTimeout.
            retry_delay (int): Time in seconds to wait between retries.

        Returns:
            None
        """
        assert self._server is not None
        assert self._target is not None
        assert max_retries >= 0
        assert retry_delay >= 0
        for retries in reversed(range(max_retries)):
            try:
                self._target.launch(location, env_mod=env_mod)
            except TargetLaunchTimeout:
                # likely has nothing to do with Grizzly but is seen frequently
                # on machines under a high load. After multiple consecutive timeouts
                # something is likely wrong so raise.
                if retries:
                    LOG.warning("Launch timeout (attempts remaining %d)", retries)
                    sleep(retry_delay)
                    continue
                raise
            break

    @staticmethod
    def location(srv_path, srv_port, close_after=None, forced_close=True, timeout=None):
        """Build a valid URL to pass to a browser.

        Args:
            srv_path (str): Path segment of the URL
            srv_port (int): Server listening port
            close_after (int): Harness argument.
            forced_close (bool): Harness argument.
            timeout (int): Harness argument.

        Returns:
            str: A valid URL.
        """
        location = "http://127.0.0.1:%d/%s" % (srv_port, srv_path.lstrip("/"))
        # set harness related arguments
        args = []
        if close_after is not None:
            assert close_after >= 0
            args.append("close_after=%d" % (close_after,))
        if not forced_close:
            args.append("forced_close=0")
        if timeout is not None:
            assert timeout >= 0
            args.append("timeout=%d" % (timeout * 1000,))
        if args:
            return "?".join([location, "&".join(args)])
        return location

    def run(self, ignore, server_map, testcase, coverage=False, wait_for_callback=False):
        """Serve a testcase and monitor the target for results.

        Args:
            ignore (list): List of failure types to ignore.
            server_map (sapphire.ServerMap): A ServerMap.
            testcase (grizzly.TestCase): The test case that will be served.
            coverage (bool): Trigger coverage dump.
            wait_for_callback: (bool): Use `_keep_waiting()` to indicate when
                                       framework should move on.

        Returns:
            None
        """
        # set initial state
        self.served = None
        self.result = None
        self.timeout = False
        if self._idle is not None:
            self._idle.schedule_poll(initial=True)
        # serve the test case
        server_status, self.served = self._server.serve_testcase(
            testcase,
            continue_cb=self._keep_waiting,
            forever=wait_for_callback,
            server_map=server_map,
            working_path=grz_tmp("serve"))
        # add all include files that were served
        for url, resource in server_map.include.items():
            testcase.add_batch(resource.target, self.served, prefix=url)
        self.timeout = server_status == SERVED_TIMEOUT
        served_lpage = testcase.landing_page in self.served
        if not served_lpage:
            LOG.debug("%r not served!", testcase.landing_page)
        elif coverage and not self.timeout:
            # dump_coverage() should be called before detect_failure()
            # to help catch any coverage related issues.
            self._target.dump_coverage()
        # detect failure
        failure_detected = self._target.detect_failure(ignore, self.timeout)
        if failure_detected == self._target.RESULT_FAILURE:
            self.result = self.FAILED
        elif not served_lpage:
            # something is wrong so close the target
            # previous iteration put target in a bad state?
            self._target.close()
            self.result = self.ERROR
        elif failure_detected == self._target.RESULT_IGNORED:
            self.result = self.IGNORED
        else:
            self.result = self.COMPLETE

    def _keep_waiting(self):
        """Callback used by the server to determine if should continue to wait
        for the requests from the target.

        Args:
            None

        Returns:
            bool: Continue to serve test test case
        """
        if self._idle is not None and self._idle.is_idle():
            LOG.debug("idle target detected")
            return False
        return self._target.monitor.is_healthy()



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""
unit tests for grizzly.replay.main
"""
from os.path import join as pathjoin
from shutil import rmtree

from pytest import raises

from ..common import TestCase, TestCaseLoadFailure
from ..target import Target, TargetLaunchError, TargetLaunchTimeout
from ..replay import ReplayManager
from ..replay.args import ReplayArgs


def test_args_01(capsys, tmp_path):
    """test parsing args"""
    # missing args tests
    with raises(SystemExit):
        ReplayArgs().parse_args([])

    # test case directory missing test_info.json
    exe = tmp_path / "binary"
    exe.touch()
    inp = tmp_path / "input"
    inp.mkdir()
    (inp / "somefile").touch()
    with raises(SystemExit):
        ReplayArgs().parse_args([str(exe), str(inp)])
    assert "error: Test case folder must contain 'test_info.json'" in capsys.readouterr()[-1]

    (inp / "test_info.json").touch()
    # specified prefs.js missing
    with raises(SystemExit):
        ReplayArgs().parse_args([str(exe), str(inp / "somefile"), "--prefs", "missing"])
    assert "error: -p/--prefs not found 'missing'" in capsys.readouterr()[-1]

    # test case directory
    (inp / "prefs.js").touch()
    ReplayArgs().parse_args([str(exe), str(inp)])

    # test case file
    ReplayArgs().parse_args([str(exe), str(inp / "somefile"), "--prefs", str(inp / "prefs.js")])

    # test negative min-crashes value
    with raises(SystemExit):
        ReplayArgs().parse_args([str(exe), str(inp), "--min-crashes", "-1"])
    assert "error: '--min-crashes' value must be positive" in capsys.readouterr()[-1]

    # test negative repeat value
    with raises(SystemExit):
        ReplayArgs().parse_args([str(exe), str(inp), "--repeat", "-1"])
    assert "error: '--repeat' value must be positive" in capsys.readouterr()[-1]

    # test missing signature file
    with raises(SystemExit):
        ReplayArgs().parse_args([str(exe), str(inp), "--sig", "missing"])
    assert "error: signature file not found" in capsys.readouterr()[-1]

    # test any crash and signature
    with raises(SystemExit):
        ReplayArgs().parse_args([str(exe), str(inp), "--any-crash", "--sig", "x"])
    assert "error: signature is ignored when running with '--any-crash'" in capsys.readouterr()[-1]

def test_main_01(mocker, tmp_path):
    """test ReplayManager.main()"""
    # This is a typical scenario - a test that reproduces results ~50% of the time.
    # Of the four attempts only the first and third will 'reproduce' the result
    # and the forth attempt should be skipped.
    # mock Sapphire.serve_testcase only
    serve_testcase = mocker.patch("grizzly.replay.replay.Sapphire.serve_testcase", autospec=True)
    serve_testcase.return_value = (None, ["test.html"])  # passed to mocked Target.detect_failure
    # setup Target
    load_target = mocker.patch("grizzly.replay.replay.load_target")
    target = mocker.Mock(spec=Target, binary="bin", forced_close=True)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.RESULT_IGNORED = Target.RESULT_IGNORED
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.side_effect = (Target.RESULT_FAILURE, Target.RESULT_NONE, Target.RESULT_FAILURE)
    def _fake_save_logs(result_logs):
        """write fake log data to disk"""
        with open(pathjoin(result_logs, "log_stderr.txt"), "w") as log_fp:
            log_fp.write("STDERR log\n")
        with open(pathjoin(result_logs, "log_stdout.txt"), "w") as log_fp:
            log_fp.write("STDOUT log\n")
        with open(pathjoin(result_logs, "log_asan_blah.txt"), "w") as log_fp:
            log_fp.write("==1==ERROR: AddressSanitizer: ")
            log_fp.write("SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n")
            log_fp.write("    #0 0xbad000 in foo /file1.c:123:234\n")
            log_fp.write("    #1 0x1337dd in bar /file2.c:1806:19\n")
    target.save_logs = _fake_save_logs
    load_target.return_value.return_value = target
    # setup args
    args = mocker.Mock()
    args.fuzzmanager = False
    args.ignore = ["fake", "timeout"]
    log_path = (tmp_path / "logs")
    args.logs = str(log_path)
    (tmp_path / "test.html").touch()
    args.input = str(tmp_path / "test.html")
    args.min_crashes = 2
    (tmp_path / "prefs.js").touch()
    args.prefs = str(tmp_path / "prefs.js")
    args.relaunch = 1
    args.repeat = 4
    (tmp_path / "sig.json").write_bytes(b"{\"symptoms\": [{\"type\": \"crashAddress\", \"address\": \"0\"}]}")
    args.sig = str(tmp_path / "sig.json")
    args.timeout = 10
    assert ReplayManager.main(args) == 0
    assert target.forced_close
    assert target.reverse.call_count == 1
    assert target.launch.call_count == 3
    assert target.step.call_count == 3
    assert target.detect_failure.call_count == 3
    assert serve_testcase.call_count == 3
    assert load_target.call_count == 1
    assert target.close.call_count == 1
    assert target.cleanup.call_count == 1
    assert target.check_relaunch.call_count == 2
    assert log_path.is_dir()
    assert any(log_path.glob('**/log_asan_blah.txt'))
    assert any(log_path.glob('**/log_stderr.txt'))
    assert any(log_path.glob('**/log_stdout.txt'))

def test_main_02(mocker):
    """test ReplayManager.main() failure cases"""
    mocker.patch("grizzly.replay.replay.FuzzManagerReporter", autospec=True)
    mocker.patch("grizzly.replay.replay.load_target", autospec=True)
    mocker.patch("grizzly.replay.replay.Sapphire", autospec=True)
    mocker.patch("grizzly.replay.replay.TestCase", autospec=True)
    # setup args
    args = mocker.Mock()
    args.ignore = None
    args.input = "test"
    args.min_crashes = 1
    args.prefs = None
    args.relaunch = 1
    args.repeat = 1
    args.sig = None

    mocker.patch("grizzly.replay.replay.ReplayManager.run", side_effect=TargetLaunchError)
    assert ReplayManager.main(args) == 1

    mocker.patch("grizzly.replay.replay.ReplayManager.run", side_effect=TargetLaunchTimeout)
    assert ReplayManager.main(args) == 1

    mocker.patch("grizzly.replay.replay.load_target", side_effect=KeyboardInterrupt)
    assert ReplayManager.main(args) == 1

    mocker.patch("grizzly.replay.replay.TestCase.load_path", side_effect=TestCaseLoadFailure)
    assert ReplayManager.main(args) == 1

def test_main_03(mocker):
    """test ReplayManager.main() loading GRZ_FORCED_CLOSE from test case"""
    mocker.patch("grizzly.replay.replay.Sapphire.serve_testcase", return_value=(None, ["x.html"]))
    target = mocker.Mock(spec=Target, forced_close=True)
    load_target = mocker.patch("grizzly.replay.replay.load_target", autospec=True)
    load_target.return_value.return_value = target
    testcase = mocker.Mock(
        spec=TestCase,
        env_vars={"GRZ_FORCED_CLOSE": "0"},
        landing_page="x.html")
    mocker.patch("grizzly.replay.replay.TestCase.load_path", return_value=testcase)
    # setup args
    args = mocker.Mock()
    args.fuzzmanager = False
    args.ignore = None
    args.input = "test"
    args.logs = None
    args.min_crashes = 1
    args.prefs = None
    args.relaunch = 1
    args.repeat = 1
    args.sig = None
    args.timeout = 1

    ReplayManager.main(args)
    assert testcase.cleanup.call_count == 1
    assert target.cleanup.call_count == 1
    assert not target.forced_close

def test_main_04(mocker, tmp_path):
    """test ReplayManager.main() loading/generating prefs.js"""
    serve_testcase = mocker.patch("grizzly.replay.replay.Sapphire.serve_testcase", autospec=True)
    serve_testcase.return_value = (None, ["test.html"])  # passed to mocked Target.detect_failure
    # setup Target
    target = mocker.Mock(spec=Target, binary="bin", forced_close=True)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    def _fake_save_logs(result_logs):
        """write fake log data to disk"""
        with open(pathjoin(result_logs, "log_stderr.txt"), "w") as log_fp:
            pass
        with open(pathjoin(result_logs, "log_stdout.txt"), "w") as log_fp:
            pass
        with open(pathjoin(result_logs, "log_asan_blah.txt"), "w") as log_fp:
            log_fp.write("==1==ERROR: AddressSanitizer: ")
            log_fp.write("SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n")
            log_fp.write("    #0 0xbad000 in foo /file1.c:123:234\n")
    target.save_logs = _fake_save_logs
    load_target = mocker.patch("grizzly.replay.replay.load_target")
    load_target.return_value.return_value = target
    # setup args
    args = mocker.Mock(
        fuzzmanager=False,
        ignore=None,
        min_crashes=1,
        relaunch=1,
        repeat=1,
        sig=None,
        timeout=1)
    log_path = (tmp_path / "logs")
    args.logs = str(log_path)
    input_path = (tmp_path / "input")
    input_path.mkdir()
    # build a test case
    entry_point = (input_path / "target.bin")
    entry_point.touch()
    with TestCase("target.bin", None, "test-adapter") as src:
        src.add_from_file(str(entry_point))
        src.dump(str(input_path), include_details=True)
    args.input = str(input_path)

    # test no specified prefs.js
    args.prefs = None
    assert ReplayManager.main(args) == 0
    assert target.launch.call_count == 1
    assert target.detect_failure.call_count == 1
    assert serve_testcase.call_count == 1
    assert log_path.is_dir()
    assert not any(log_path.glob('**/prefs.js'))

    target.reset_mock()
    serve_testcase.reset_mock()
    rmtree(str(log_path), ignore_errors=True)

    # test included prefs.js
    (input_path / "prefs.js").write_bytes(b"included")
    assert ReplayManager.main(args) == 0
    assert target.launch.call_count == 1
    assert target.detect_failure.call_count == 1
    assert serve_testcase.call_count == 1
    assert log_path.is_dir()
    prefs = tuple(log_path.glob('**/prefs.js'))
    assert prefs[0].read_bytes() == b"included"

    target.reset_mock()
    serve_testcase.reset_mock()
    rmtree(str(log_path), ignore_errors=True)

    # test specified prefs.js
    (tmp_path / "prefs.js").write_bytes(b"specified")
    args.prefs = str(tmp_path / "prefs.js")
    assert ReplayManager.main(args) == 0
    assert target.launch.call_count == 1
    assert target.detect_failure.call_count == 1
    assert serve_testcase.call_count == 1
    assert log_path.is_dir()
    prefs = tuple(log_path.glob('**/prefs.js'))
    assert prefs[0].read_bytes() == b"specified"



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .utils import grz_tmp

def test_testcase_01(mocker, tmp_path):
    """test grz_tmp()"""
    mocker.patch("grizzly.common.utils.gettempdir", autospec=True, return_value=str(tmp_path))
    # create temp path
    path = grz_tmp()
    assert path == str(tmp_path / "grizzly")
    assert (tmp_path / "grizzly").is_dir()
    # create temp path (exists)
    path = grz_tmp()
    assert path == str(tmp_path / "grizzly")
    # create temp path with sub directory
    path = grz_tmp("test1", "test2")
    assert path == str(tmp_path / "grizzly" / "test1" / "test2")
    assert (tmp_path / "grizzly" / "test1" / "test2").is_dir()



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""
unit tests for grizzly.replay
"""
from os.path import join as pathjoin

from pytest import raises

from sapphire import Sapphire, SERVED_ALL, SERVED_REQUEST
from .replay import ReplayManager
from ..common import Report, Status, TestCase
from ..target import Target, TargetLaunchError


def _fake_save_logs_result(result_logs, meta=False):  # pylint: disable=unused-argument
    """write fake log data to disk"""
    with open(pathjoin(result_logs, "log_stderr.txt"), "w") as log_fp:
        log_fp.write("STDERR log\n")
    with open(pathjoin(result_logs, "log_stdout.txt"), "w") as log_fp:
        log_fp.write("STDOUT log\n")
    with open(pathjoin(result_logs, "log_asan_blah.txt"), "w") as log_fp:
        log_fp.write("==1==ERROR: AddressSanitizer: ")
        log_fp.write("SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n")
        log_fp.write("    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write("    #1 0x1337dd in bar /file2.c:1806:19\n")

def test_replay_01(mocker):
    """test ReplayManager.cleanup()"""
    replay = ReplayManager([], mocker.Mock(spec=Sapphire), mocker.Mock(spec=Target), mocker.Mock())
    replay._reports_expected = {"A":  mocker.Mock(spec=Report)}
    replay._reports_other = {"B":  mocker.Mock(spec=Report)}
    replay.status = mocker.Mock(spec=Status)
    ereport = tuple(replay.reports)[0]
    oreport = tuple(replay.other_reports)[0]
    replay.cleanup()
    assert ereport.cleanup.call_count == 1
    assert oreport.cleanup.call_count == 1
    assert replay.status.cleanup.call_count == 1

def test_replay_02(mocker):
    """test ReplayManager.run() - no repro"""
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_testcase.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target)
    target.RESULT_NONE = Target.RESULT_NONE
    target.closed = True
    target.detect_failure.return_value = Target.RESULT_NONE
    target.forced_close = True
    target.rl_reset = 1
    testcase = mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html")
    with ReplayManager([], server, target, testcase, use_harness=True) as replay:
        assert not replay.run()
        assert replay.status.ignored == 0
        assert replay.status.iteration == 1
        assert replay.status.results == 0
        assert not replay.reports

def test_replay_03(mocker):
    """test ReplayManager.run() - successful repro"""
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_testcase.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.binary = "C:\\fake_bin"
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.save_logs = _fake_save_logs_result
    testcase = mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html")
    with ReplayManager([], server, target, testcase, use_harness=False) as replay:
        assert replay.run()
        assert replay.status.ignored == 0
        assert replay.status.iteration == 1
        assert replay.status.results == 1
        assert len(replay.reports) == 1
        assert not replay.other_reports

def test_replay_04(mocker):
    """test ReplayManager.run() - Error (landing page not requested/served)"""
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_testcase.return_value = (SERVED_REQUEST, ["x"])
    target = mocker.Mock(spec=Target)
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.return_value = Target.RESULT_NONE
    testcase = mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html")
    with ReplayManager([], server, target, testcase, use_harness=False) as replay:
        assert not replay.run(repeat=2)
        assert replay.status.ignored == 0
        assert replay.status.iteration == 1
        assert replay.status.results == 0
        assert not replay.reports
        assert not replay.other_reports
    assert target.check_relaunch.call_count == 0

def test_replay_05(mocker):
    """test ReplayManager.run() - ignored"""
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_testcase.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target)
    target.RESULT_IGNORED = Target.RESULT_IGNORED
    target.detect_failure.return_value = Target.RESULT_IGNORED
    testcase = mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html")
    with ReplayManager([], server, target, testcase, use_harness=False) as replay:
        assert not replay.run()
        assert replay.status.ignored == 1
        assert replay.status.iteration == 1
        assert replay.status.results == 0
        assert not replay.reports
        assert not replay.other_reports

def test_replay_06(mocker):
    """test ReplayManager.run() - early exit"""
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_testcase.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.RESULT_IGNORED = Target.RESULT_IGNORED
    target.RESULT_NONE = Target.RESULT_NONE
    target.binary = "path/fake_bin"
    target.save_logs = _fake_save_logs_result
    testcase = mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html")
    # early failure
    target.detect_failure.side_effect = [Target.RESULT_FAILURE, Target.RESULT_IGNORED, Target.RESULT_NONE]
    with ReplayManager([], server, target, testcase, use_harness=False) as replay:
        assert not replay.run(repeat=4, min_results=3)
        assert replay.status.iteration == 3
        assert replay.status.results == 1
        assert replay.status.ignored == 1
        assert len(replay.reports) == 1
    # early success
    target.detect_failure.side_effect = [Target.RESULT_FAILURE, Target.RESULT_IGNORED, Target.RESULT_FAILURE]
    with ReplayManager([], server, target, testcase, use_harness=False) as replay:
        assert replay.run(repeat=4, min_results=2)
        assert replay.status.iteration == 3
        assert replay.status.results == 2
        assert replay.status.ignored == 1
        assert len(replay._reports_expected) == 1
        assert not replay._reports_other
        assert len(replay.reports) == 1

def test_replay_07(mocker, tmp_path):
    """test ReplayManager.run() - test signatures"""
    report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    mocker.patch("grizzly.replay.replay.mkdtemp", autospec=True, return_value=str(tmp_path))
    report_0 = mocker.Mock(spec=Report)
    report_0.crash_info.return_value.createShortSignature.return_value = "No crash detected"
    report_1 = mocker.Mock(spec=Report)
    report_1.crash_info.return_value.createShortSignature.return_value = "[@ test1]"
    report_1.major = "0123abcd"
    report_1.minor = "01239999"
    report_2 = mocker.Mock(spec=Report)
    report_2.crash_info.return_value.createShortSignature.return_value = "[@ test2]"
    report_2.major = "0123abcd"
    report_2.minor = "abcd9876"
    report.from_path.side_effect = (report_0, report_1, report_2)
    server = mocker.Mock(spec=Sapphire)
    server.port = 0x1337
    server.serve_testcase.return_value = (SERVED_ALL, ["index.html"])
    signature = mocker.Mock()
    signature.matches.side_effect = (True, False)
    target = mocker.Mock(spec=Target)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.binary = "fake_bin"
    testcase = mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html")
    with ReplayManager([], server, target, testcase, signature=signature, use_harness=False) as replay:
        assert not replay.run(repeat=3, min_results=2)
        assert replay._signature == signature
        assert report.from_path.call_count == 3
        assert replay.status.iteration == 3
        assert replay.status.results == 1
        assert replay.status.ignored == 1
        assert len(replay.reports) == 1
        assert len(replay.other_reports) == 1
        assert report_0.cleanup.call_count == 1
        assert report_1.cleanup.call_count == 0
        assert report_2.cleanup.call_count == 0
    assert signature.matches.call_count == 2

def test_replay_08(mocker, tmp_path):
    """test ReplayManager.run() - any crash"""
    report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    mocker.patch("grizzly.replay.replay.mkdtemp", autospec=True, return_value=str(tmp_path))
    report_0 = mocker.Mock(spec=Report)
    report_0.crash_info.return_value.createShortSignature.return_value = "No crash detected"
    report_1 = mocker.Mock(spec=Report)
    report_1.crash_info.return_value.createShortSignature.return_value = "[@ test1]"
    report_1.crash_hash.return_value = "hash1"
    report_1.major = "0123abcd"
    report_1.minor = "01239999"
    report_2 = mocker.Mock(spec=Report)
    report_2.crash_info.return_value.createShortSignature.return_value = "[@ test2]"
    report_2.crash_hash.return_value = "hash2"
    report_2.major = "0123abcd"
    report_2.minor = "abcd9876"
    report.from_path.side_effect = (report_0, report_1, report_2)
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_testcase.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.binary = "fake_bin"
    testcase = mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html")
    with ReplayManager([], server, target, testcase, any_crash=True, use_harness=False) as replay:
        assert replay.run(repeat=3, min_results=2)
        assert replay._signature is None
        assert report.from_path.call_count == 3
        assert replay.status.iteration == 3
        assert replay.status.results == 2
        assert replay.status.ignored == 0
        assert report_1.crash_hash.call_count == 1
        assert report_2.crash_hash.call_count == 1
        assert len(replay.reports) == 2
        assert not replay.other_reports
        assert report_0.cleanup.call_count == 1
        assert report_1.cleanup.call_count == 0
        assert report_2.cleanup.call_count == 0

def test_replay_09(mocker, tmp_path):
    """test ReplayManager.report_to_filesystem()"""
    # no reports
    ReplayManager.report_to_filesystem(str(tmp_path), [])
    assert not any(tmp_path.glob("*"))
    # with reports
    reports_expected = list()
    reports_expected.append(mocker.Mock(spec=Report))
    reports_expected[-1].prefix = "expected"
    (tmp_path / "report_expected").mkdir()
    reports_expected[-1].path = str(tmp_path / "report_expected")
    reports_other = list()
    reports_other.append(mocker.Mock(spec=Report))
    reports_other[-1].prefix = "other1"
    (tmp_path / "report_other1").mkdir()
    reports_other[-1].path = str(tmp_path / "report_other1")
    reports_other.append(mocker.Mock(spec=Report))
    reports_other[-1].prefix = "other2"
    (tmp_path / "report_other2").mkdir()
    reports_other[-1].path = str(tmp_path / "report_other2")
    test = mocker.Mock(spec=TestCase)
    path = tmp_path / "dest"
    ReplayManager.report_to_filesystem(str(path), reports_expected, reports_other, test=test)
    assert test.dump.call_count == 3  # called once per report
    assert not (tmp_path / "report_expected").is_dir()
    assert not (tmp_path / "report_other1").is_dir()
    assert not (tmp_path / "report_other2").is_dir()
    assert path.is_dir()
    assert (path / "reports").is_dir()
    assert (path / "reports" / "expected_logs").is_dir()
    assert (path / "other_reports").is_dir()
    assert (path / "other_reports" / "other1_logs").is_dir()
    assert (path / "other_reports" / "other2_logs").is_dir()

def test_replay_10(mocker, tmp_path):
    """test ReplayManager.run() - TargetLaunchError"""
    report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    mocker.patch("grizzly.replay.replay.mkdtemp", autospec=True, return_value=str(tmp_path))
    report.from_path.side_effect = (mocker.Mock(spec=Report),)
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    target = mocker.Mock(spec=Target)
    target.launch.side_effect = TargetLaunchError
    testcase = mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html")
    with ReplayManager([], server, target, testcase, use_harness=False) as replay:
        with raises(TargetLaunchError):
            replay.run()
        assert not any(replay.reports)
        assert any(replay.other_reports)
        assert "STARTUP" in replay._reports_other



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly status reporter"""
# pylint: disable=protected-access

import re

import pytest

from .status_reporter import main, ReducerStats, Status, StatusReporter, TracebackReport

def _fake_sys_info():
    return "CPU & Load : 64 @ 93.1% (85.25, 76.21, 51.06)\n" \
           "    Memory : 183.9GB of 251.9GB free\n" \
           "      Disk : 22.2GB of 28.7GB free"

def test_status_reporter_01(tmp_path):
    """test basic StatusReporter"""
    st_rpt = StatusReporter(list())
    st_rpt._sys_info = _fake_sys_info
    out = st_rpt._specific()
    assert "No status reports available" in out
    report = tmp_path / "output.txt"
    st_rpt.dump_specific(str(report))
    assert report.is_file()
    st_rpt.print_specific()
    out = st_rpt._summary()
    assert "No status reports available" in out
    report.unlink()
    st_rpt.dump_summary(str(report))
    assert report.is_file()
    st_rpt.print_summary()

def test_status_reporter_02(tmp_path):
    """test StatusReporter.load()"""
    Status.PATH = str(tmp_path / "grzstatus")
    with pytest.raises(IOError):
        StatusReporter.load(tb_path="no_dir")
    st_rpt = StatusReporter.load(tb_path=str(tmp_path))
    assert isinstance(st_rpt.reports, list)
    assert not st_rpt.reports
    assert isinstance(st_rpt.tracebacks, list)
    assert not st_rpt.tracebacks

def test_status_reporter_03(mocker):
    """test StatusReporter._sys_info()"""
    gbs = 1024 * 1024 * 1024
    fake_psutil = mocker.patch("grizzly.common.status_reporter.psutil", autospec=True)
    fake_psutil.cpu_count.return_value = 4
    fake_psutil.cpu_percent.return_value = 10.0
    fake_psutil.virtual_memory.return_value = mocker.Mock(available=12, total=gbs)
    fake_psutil.disk_usage.return_value = mocker.Mock(free=12, total=gbs)
    sysinfo = StatusReporter._sys_info()
    assert "MB" in sysinfo
    fake_psutil.virtual_memory.return_value = mocker.Mock(available=1.1 * gbs, total=2 * gbs)
    fake_psutil.disk_usage.return_value = mocker.Mock(free=10.23 * gbs, total=100 * gbs)
    sysinfo = StatusReporter._sys_info()
    assert "MB" not in sysinfo
    lines = sysinfo.split("\n")
    assert len(lines) == 3
    assert "CPU & Load : " in lines[0]
    assert "Memory : " in lines[1]
    assert "Disk : " in lines[2]
    # verify alignment
    position = len(lines[0].split(":")[0])
    for line in lines:
        assert line[position] == ":"

def test_status_reporter_04(tmp_path):
    """test StatusReporter._scan()"""
    re_filter = re.compile("TEST_FILE")
    (tmp_path / "somefile.txt").touch()
    test_path = tmp_path / "TEST_FILE"
    test_path.mkdir()
    assert not any(StatusReporter._scan(str(tmp_path), re_filter))
    test_path.rmdir()
    test_path.touch()
    assert not any(StatusReporter._scan(str(tmp_path), re_filter))
    test_path.write_bytes(b"test")
    assert tuple(StatusReporter._scan(str(tmp_path), re_filter))

def test_status_reporter_05(tmp_path):
    """test StatusReporter._summary()"""
    Status.PATH = str(tmp_path / "grzstatus")
    # single report
    status = Status.start()
    status.ignored = 0
    status.iteration = 1
    status.log_size = 0
    status.results = 0
    status.report(force=True)
    rptr = StatusReporter.load(tb_path=str(tmp_path))
    rptr._sys_info = _fake_sys_info
    assert rptr.reports is not None
    assert len(rptr.reports) == 1
    output = rptr._summary(runtime=False)
    assert "Iteration" in output
    assert "Rate" in output
    assert "Results" in output
    assert "ignored" not in output
    assert "Logs" not in output
    assert "Runtime" not in output
    assert "Timestamp" not in output
    assert len(output.split("\n")) == 3
    # multiple reports
    status = Status.start()
    status.start_time += 66.0
    status.ignored = 1
    status.iteration = 8
    status.log_size = 86900000
    status.results = 0
    status.report(force=True)
    rptr = StatusReporter.load(tb_path=str(tmp_path))
    rptr._sys_info = _fake_sys_info
    assert len(rptr.reports) == 2
    output = rptr._summary(sysinfo=True, timestamp=True)
    assert "Iteration" in output
    assert "Rate" in output
    assert "Results" in output
    assert "ignored" in output
    assert "Logs" in output
    assert "Runtime" in output
    assert "Timestamp" in output
    lines = output.split("\n")
    assert len(lines) == 9
    # verify alignment
    position = len(lines[0].split(":")[0])
    for line in lines:
        assert re.match(r"\S\s:\s\S", line[position - 2:])

def test_status_reporter_06(tmp_path):
    """test StatusReporter._specific()"""
    Status.PATH = str(tmp_path / "grzstatus")
    # single report
    status = Status.start()
    status.ignored = 0
    status.iteration = 1
    status.log_size = 0
    status.results = 0
    status.report(force=True)
    rptr = StatusReporter.load()
    rptr._sys_info = _fake_sys_info
    assert rptr.reports is not None
    output = rptr._specific()
    assert len(output.split("\n")[:-1]) == 2
    assert "Ignored" in output
    assert "Iteration" in output
    assert "Rate" in output
    assert "Results" in output
    assert "EXPIRED" not in output
    # multiple reports
    status = Status.start()
    status.ignored = 1
    status.iteration = 432422
    status.results = 123
    status.report(force=True)
    rptr = StatusReporter.load()
    rptr._sys_info = _fake_sys_info
    assert len(rptr.reports) == 2
    output = rptr._specific()
    assert len(output.split("\n")[:-1]) == 4
    assert "Ignored" in output
    assert "Iteration" in output
    assert "Rate" in output
    assert "Results" in output

def test_status_reporter_07(tmp_path):
    """test StatusReporter.load() with traceback"""
    Status.PATH = str(tmp_path / "grzstatus")
    status = Status.start()
    status.ignored = 0
    status.iteration = 1
    status.log_size = 0
    status.results = 0
    status.report(force=True)
    # create boring screenlog
    (tmp_path / "screenlog.0").write_bytes(b"boring\ntest\n123\n")
    # create first screenlog
    with (tmp_path / "screenlog.1").open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  blah\n")
        test_fp.write(b"IndexError: list index out of range\n")
    rptr = StatusReporter.load(tb_path=str(tmp_path))
    assert len(rptr.tracebacks) == 1
    # create second screenlog
    with (tmp_path / "screenlog.1234").open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  blah\n")
        test_fp.write(b"foo.bar.error: blah\n")
    rptr = StatusReporter.load(tb_path=str(tmp_path))
    assert len(rptr.tracebacks) == 2
    # create third screenlog
    with (tmp_path / "screenlog.3").open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  blah\n")
        test_fp.write(b"KeyboardInterrupt\n")
    rptr = StatusReporter.load(tb_path=str(tmp_path))
    assert len(rptr.tracebacks) == 2
    merged_log = rptr._summary()
    assert len(merged_log.splitlines()) == 14
    assert "screenlog.1" in merged_log
    assert "screenlog.1234" in merged_log
    assert "IndexError" in merged_log
    assert "foo.bar.error" in merged_log
    assert "screenlog.3" not in merged_log

def test_status_reporter_08(tmp_path):
    """test StatusReporter.load() no reports with traceback"""
    Status.PATH = str(tmp_path / "grzstatus")
    # create screenlog with tb
    with (tmp_path / "screenlog.1").open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  blah\n")
        test_fp.write(b"IndexError: list index out of range\n")
    rptr = StatusReporter.load(tb_path=str(tmp_path))
    rptr._sys_info = _fake_sys_info
    assert len(rptr.tracebacks) == 1
    output = rptr._summary()
    assert len(output.splitlines()) == 7
    assert "No status reports available" in output
    assert "IndexError" in output

def test_status_reporter_09(tmp_path):
    """test StatusReporter.summary() limit with traceback"""
    Status.PATH = str(tmp_path / "grzstatus")
    # create reports
    status = Status.start()
    status.ignored = 100
    status.iteration = 1000
    status.log_size = 9999999999
    status.results = 123
    status.report(force=True)
    status = Status.start()
    status.ignored = 9
    status.iteration = 192938
    status.log_size = 0
    status.results = 3
    status.report(force=True)
    # create screenlogs with tracebacks
    for i in range(10):
        with (tmp_path / ("screenlog.%d" % (i,))).open("wb") as test_fp:
            test_fp.write(b"Traceback (most recent call last):\n")
            for j in range(TracebackReport.MAX_LINES):
                test_fp.write(b"  File \"some/long/path/name/foobar.py\", line 5000, in <module>\n")
                test_fp.write(b"    some_long_name_for_a_func_%04d()\n" % (j,))
            test_fp.write(b"IndexError: list index out of range\n")
    rptr = StatusReporter.load(tb_path=str(tmp_path))
    rptr._sys_info = _fake_sys_info
    assert len(rptr.tracebacks) == 10
    merged_log = rptr._summary(runtime=True, sysinfo=True, timestamp=True)
    assert len(merged_log) < StatusReporter.SUMMARY_LIMIT

def test_reduce_status_reporter_01(tmp_path):
    """test empty StatusReporter in reducer mode"""
    Status.PATH = str(tmp_path / "grzstatus")
    ReducerStats.PATH = str(tmp_path)
    rptr = StatusReporter.load(reducer=True)
    assert rptr is not None
    assert not rptr.reports
    output = rptr._specific()
    assert "No status reports available" in output
    output = rptr._summary()
    assert "No status reports available" in output

def test_reduce_status_reporter_02(tmp_path):
    """test StatusReporter._specific() in reducer mode"""
    Status.PATH = str(tmp_path / "grzstatus")
    ReducerStats.PATH = str(tmp_path)
    status = Status.start()
    status.ignored = 12
    status.iteration = 432422
    status.results = 123
    status.report(force=True)
    rptr = StatusReporter.load(reducer=True)
    rptr._sys_info = _fake_sys_info
    assert rptr.reports
    output = rptr._specific()
    assert len(output.split("\n")[:-1]) == 2
    assert "Iteration" in output
    assert "Rate" in output
    assert "Ignored" not in output
    assert "Results" not in output

def test_reduce_status_reporter_03(tmp_path):
    """test StatusReporter._summary() in reducer mode"""
    Status.PATH = str(tmp_path / "grzstatus")
    ReducerStats.PATH = str(tmp_path)
    status = Status.start()
    status.iteration = 1
    status.report(force=True)
    status = Status.start()
    status.iteration = 10
    status.report(force=True)
    rptr = StatusReporter.load(reducer=True)
    rptr._sys_info = _fake_sys_info
    assert rptr.reports is not None
    output = rptr._summary(sysinfo=True, timestamp=True)
    assert "======== Stats ========" in output
    assert "======= Active ========" in output
    assert "Reduced" in output
    assert "No Repro" in output
    assert "Iteration" in output
    assert "Results" not in output
    assert "Mismatch" in output
    assert "Rate" in output
    assert "Runtime" in output
    assert "ignored" not in output
    assert "Timestamp" in output
    lines = output.split("\n")
    assert len(lines) == 13
    # verify alignment
    position = len(lines[1].split(":")[0])
    for line in lines:
        if line.startswith("="):
            # skip headers
            continue
        assert re.match(r"\S\s:\s\S", line[position - 2:])

def test_traceback_report_01():
    """test simple TracebackReport"""
    tbr = TracebackReport("log.txt", ["0", "1", "2"], prev_lines=["-2", "-1"])
    output = str(tbr)
    assert len(output.splitlines()) == 6
    assert len(tbr) == 26
    assert "log.txt" in output
    assert "2" in output
    assert "-2" in output

def test_traceback_report_02():
    """test empty TracebackReport"""
    tbr = TracebackReport("log.txt", [])
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 1
    assert "log.txt" in output

def test_traceback_report_03(tmp_path):
    """test TracebackReport.from_file()"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("wb") as test_fp:
        test_fp.write(b"start junk\npre1\npre2\npre3\npre4\npre5\n")
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  File \"foo.py\", line 556, in <module>\n")
        test_fp.write(b"    main(parse_args())\n")
        test_fp.write(b"  File \"foo.py\", line 207, in bar\n")
        test_fp.write(b"    a = b[10]\n")
        test_fp.write(b"IndexError: list index out of range\n")
        test_fp.write(b"end junk\n")
    tbr = TracebackReport.from_file(str(test_log))
    assert len(tbr.prev_lines) == 5
    assert len(tbr.lines) == 6
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 12
    assert "pre1" in output
    assert "IndexError" in output
    assert "screenlog.0" in output
    assert "junk" not in output

    with test_log.open("wb") as test_fp:
        test_fp.write(b"start junk\n")
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  File \"foo.py\", line 556, in <module>\n")
        test_fp.write(b"    main(parse_args())\n")
        test_fp.write(b"  File \"foo.py\", line 207, in bar\n")
        test_fp.write(b"    a = b[10]\n")
        test_fp.write(b"foo.bar.error: blah\n")
        test_fp.write(b"end junk\n")
    tbr = TracebackReport.from_file(str(test_log), max_preceeding=0)
    assert len(tbr.lines) == 6
    assert not tbr.prev_lines
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 7
    assert "screenlog.0" in output
    assert "foo.bar.error" in output
    assert "junk" not in output
    # kbi
    with test_log.open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  File \"foo.py\", line 556, in <module>\n")
        test_fp.write(b"    main(parse_args())\n")
        test_fp.write(b"  File \"foo.py\", line 207, in bar\n")
        test_fp.write(b"    a = b[10]\n")
        test_fp.write(b"KeyboardInterrupt\n")
        test_fp.write(b"end junk\n")
    tbr = TracebackReport.from_file(str(test_log))
    assert tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 7
    assert "KeyboardInterrupt" in output

def test_traceback_report_04(tmp_path):
    """test TracebackReport.from_file() exceed size limit"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  File \"foo.py\", line 5, in <module>\n")
        test_fp.write(b"    first()\n")
        test_fp.write(b"  File \"foo.py\", line 5, in <module>\n")
        test_fp.write(b"    second()\n")
        for i in reversed(range(TracebackReport.MAX_LINES)):
            test_fp.write(b"  File \"foo.py\", line 5, in <module>\n")
            test_fp.write(b"    func_%02d()\n" % i)
        test_fp.write(b"END_WITH_BLANK_LINE\n\n")
        test_fp.write(b"end junk\n")
    tbr = TracebackReport.from_file(str(test_log))
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 18
    assert "<--- TRACEBACK TRIMMED--->" in output
    assert "first()" in output
    assert "func_05()" in output
    assert "second()" not in output
    assert "func_06()" not in output
    assert "END_WITH_BLANK_LINE" in output

def test_traceback_report_05(tmp_path):
    """test TracebackReport.from_file() cut off"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  File \"foo.py\", line 5, in <module>\n")
        test_fp.write(b"    first()\n")
        for i in range(TracebackReport.MAX_LINES * 2):
            test_fp.write(b"  File \"foo.py\", line 5, in <module>\n")
            test_fp.write(b"    func_%d()\n" % i)
    tbr = TracebackReport.from_file(str(test_log))
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 18
    assert "first()" in output
    assert "func_%d" % (TracebackReport.MAX_LINES * 2 - 1) in output

def test_traceback_report_06(tmp_path):
    """test TracebackReport.from_file() single word error"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  File \"foo.py\", line 5, in <module>\n")
        test_fp.write(b"    first()\n")
        test_fp.write(b"AssertionError\n")
        test_fp.write(b"end junk\n")
    tbr = TracebackReport.from_file(str(test_log))
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 5
    assert "first()" in output
    assert "AssertionError" in output
    assert "end junk" not in output

def test_traceback_report_07(tmp_path):
    """test TracebackReport.from_file() with binary data"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  File \"foo.py\", line 5, in <module>\n")
        test_fp.write(b"    bin\xd8()\n")
        test_fp.write(b"AssertionError\n")
    tbr = TracebackReport.from_file(str(test_log))
    assert not tbr.is_kbi
    output = str(tbr)
    assert "bin()" in output
    assert "AssertionError" in output

def test_traceback_report_08(tmp_path):
    """test TracebackReport.from_file() locate token across chunks"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("wb") as test_fp:
        test_fp.write(b"A" * (TracebackReport.READ_LIMIT - 5))
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  File \"foo.py\", line 5, in <module>\n")
        test_fp.write(b"    first()\n")
        test_fp.write(b"AssertionError\n")
    tbr = TracebackReport.from_file(str(test_log))
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 5
    assert "first()" in output
    assert "AssertionError" in output

def test_main_01(tmp_path):
    """test main() with no reports"""
    Status.PATH = str(tmp_path)
    StatusReporter.CPU_POLL_INTERVAL = 0.01
    assert main([]) == 0

def test_main_02(tmp_path):
    """test main() with a report"""
    Status.PATH = str(tmp_path)
    StatusReporter.CPU_POLL_INTERVAL = 0.01
    status = Status.start()
    status.iteration = 1
    status.report(force=True)
    assert main([]) == 0

def test_main_03(tmp_path):
    """test main() --dump"""
    Status.PATH = str(tmp_path)
    StatusReporter.CPU_POLL_INTERVAL = 0.01
    status = Status.start()
    status.iteration = 1
    status.report(force=True)
    dump_file = tmp_path / "output.txt"
    assert main(["--dump", str(dump_file)]) == 0
    assert dump_file.is_file()
    assert b"Runtime" not in dump_file.read_bytes()
    #assert False, dump_file.read_bytes()
    dump_file.unlink()
    dump_file = tmp_path / "output.txt"
    assert main(["--dump", str(dump_file), "--mode", "reduce-status"]) == 0
    assert dump_file.is_file()
    assert b"Runtime" in dump_file.read_bytes()

def test_main_04(tmp_path):
    """test main() with --mode reduce-status"""
    Status.PATH = str(tmp_path)
    StatusReporter.CPU_POLL_INTERVAL = 0.01
    status = Status.start()
    status.iteration = 1
    status.report(force=True)
    assert main(["--mode", "reduce-status"]) == 0

def test_main_05():
    """test main() with invalid mode"""
    with pytest.raises(SystemExit):
        main(["--mode", "invalid"])



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from os.path import join as pathjoin

from pytest import raises

from sapphire import Sapphire, SERVED_ALL, SERVED_NONE, SERVED_REQUEST, SERVED_TIMEOUT, ServerMap

from .runner import _IdleChecker, Runner
from .storage import TestCase
from ..target import Target, TargetLaunchError, TargetLaunchTimeout

def test_runner_01(mocker):
    """test Runner()"""
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)
    target.detect_failure.return_value = target.RESULT_NONE
    runner = Runner(server, target)
    assert runner._idle is None
    assert runner.result is None
    assert runner.served is None
    assert not runner.timeout
    serv_files = ["a.bin", "/another/file.bin"]
    testcase = mocker.Mock(spec=TestCase)
    testcase.landing_page = serv_files[0]
    # all files served
    server.serve_testcase.return_value = (SERVED_ALL, serv_files)
    runner.run([], ServerMap(), testcase)
    assert runner.result == runner.COMPLETE
    assert runner.served == serv_files
    assert not runner.timeout
    assert target.close.call_count == 0
    assert target.dump_coverage.call_count == 0
    # some files served
    server.serve_testcase.return_value = (SERVED_REQUEST, serv_files)
    runner.run([], ServerMap(), testcase, coverage=True)
    assert runner.result == runner.COMPLETE
    assert runner.served == serv_files
    assert not runner.timeout
    assert target.close.call_count == 0
    assert target.dump_coverage.call_count == 1

def test_runner_02(mocker):
    """test Runner() errors"""
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)
    testcase = mocker.Mock(spec=TestCase, landing_page="x")
    runner = Runner(server, target)
    # no files served
    server.serve_testcase.return_value = (SERVED_NONE, [])
    target.detect_failure.return_value = target.RESULT_NONE
    runner.run([], ServerMap(), testcase)
    assert runner.result == runner.ERROR
    assert not runner.served
    assert not runner.timeout
    assert target.close.call_count == 1
    target.reset_mock()
    # landing page not served
    server.serve_testcase.return_value = (SERVED_REQUEST, ["harness"])
    runner.run([], ServerMap(), testcase)
    assert runner.result == runner.ERROR
    assert runner.served
    assert target.close.call_count == 1

def test_runner_03(mocker):
    """test reporting timeout"""
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)
    serv_files = ["a.bin", "/another/file.bin"]
    server.serve_testcase.return_value = (SERVED_TIMEOUT, serv_files)
    runner = Runner(server, target)
    target.detect_failure.return_value = target.RESULT_FAILURE
    runner.run([], ServerMap(), mocker.Mock(spec=TestCase, landing_page="x"))
    assert runner.result == runner.FAILED
    assert runner.served == serv_files
    assert runner.timeout

def test_runner_04(mocker):
    """test reporting failures"""
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)
    serv_files = ["file.bin"]
    server.serve_testcase.return_value = (SERVED_REQUEST, serv_files)
    testcase = mocker.Mock(spec=TestCase, landing_page=serv_files[0])
    runner = Runner(server, target)
    # test FAILURE
    target.detect_failure.return_value = target.RESULT_FAILURE
    runner.run([], ServerMap(), testcase)
    assert runner.result == runner.FAILED
    assert runner.served == serv_files
    assert not runner.timeout
    # test IGNORED
    target.detect_failure.return_value = target.RESULT_IGNORED
    runner.run([], ServerMap(), testcase)
    assert runner.result == runner.IGNORED
    assert runner.served == serv_files
    assert not runner.timeout
    # failure before serving landing page
    server.serve_testcase.return_value = (SERVED_REQUEST, ["harness"])
    target.detect_failure.return_value = target.RESULT_FAILURE
    runner.run([], ServerMap(), testcase)
    assert runner.result == runner.FAILED
    assert runner.served
    assert not runner.timeout

def test_runner_05(mocker):
    """test Runner() with idle checking"""
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)
    target.detect_failure.return_value = target.RESULT_NONE
    serv_files = ["/fake/file", "/another/file.bin"]
    server.serve_testcase.return_value = (SERVED_REQUEST, serv_files)
    runner = Runner(server, target, idle_threshold=0.01, idle_delay=0.01)
    assert runner._idle is not None
    runner.run([], ServerMap(), mocker.Mock(spec=TestCase, landing_page=serv_files[0]))
    assert runner.result == runner.COMPLETE
    assert target.close.call_count == 0

def test_runner_06(mocker):
    """test Runner._keep_waiting()"""
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)

    target.monitor.is_healthy.return_value = True
    runner = Runner(server, target)
    assert runner._keep_waiting()

    target.monitor.is_healthy.return_value = False
    assert not runner._keep_waiting()

    runner._idle = mocker.Mock(spec=_IdleChecker)
    runner._idle.is_idle.return_value = False
    target.monitor.is_healthy.return_value = True
    assert runner._keep_waiting()

    runner._idle.is_idle.return_value = True
    target.monitor.is_healthy.return_value = True
    assert not runner._keep_waiting()

    runner._idle.is_idle.return_value = False
    target.monitor.is_healthy.return_value = False
    assert not runner._keep_waiting()

def test_runner_07():
    """test Runner.location()"""
    result = Runner.location("a.html", 34567)
    assert result == "http://127.0.0.1:34567/a.html"
    result = Runner.location("/a.html", 34567)
    assert result == "http://127.0.0.1:34567/a.html"
    result = Runner.location("a.html", 34567, close_after=10)
    assert result == "http://127.0.0.1:34567/a.html?close_after=10"
    result = Runner.location("a.html", 34567, close_after=10, forced_close=False)
    assert result == "http://127.0.0.1:34567/a.html?close_after=10&forced_close=0"
    result = Runner.location("a.html", 34567, forced_close=False)
    assert result == "http://127.0.0.1:34567/a.html?forced_close=0"
    result = Runner.location("a.html", 9999, close_after=10, forced_close=False, timeout=60)
    assert result == "http://127.0.0.1:9999/a.html?close_after=10&forced_close=0&timeout=60000"

def test_runner_08(mocker):
    """test Runner.launch()"""
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    target = mocker.Mock(spec=Target)

    runner = Runner(server, target)
    runner.launch("http://a/")
    assert target.launch.call_count == 1
    target.reset_mock()

    target.launch.side_effect = TargetLaunchError
    with raises(TargetLaunchError):
        runner.launch("http://a/")
    assert target.launch.call_count == 1
    target.reset_mock()

    target.launch.side_effect = TargetLaunchTimeout
    with raises(TargetLaunchTimeout):
        runner.launch("http://a/", max_retries=3)
    assert target.launch.call_count == 3

def test_runner_09(mocker, tmp_path):
    """test Runner.run() adding includes to testcase"""
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)
    target.detect_failure.return_value = target.RESULT_NONE
    runner = Runner(server, target)
    # create test files
    inc_path1 = (tmp_path / "include")
    inc_path1.mkdir()
    inc1 = (inc_path1 / "inc_file.bin")
    inc1.write_bytes(b"a")
    (inc_path1 / "nested").mkdir()
    inc2 = (inc_path1 /  "nested" / "nested_inc.bin")
    inc2.write_bytes(b"a")
    inc_path2 = (tmp_path / "include2")
    inc_path2.mkdir()
    inc3 = (inc_path2 / "inc_file3.txt")
    inc3.write_bytes(b"a")
    # build server map
    smap = ServerMap()
    smap.set_include("/", str(inc_path1))
    smap.set_include("/test", str(inc_path2))
    serv_files = ["a.b", str(inc1), str(inc2), str(inc3)]
    server.serve_testcase.return_value = (SERVED_ALL, serv_files)
    with TestCase("a.b", "x", "x") as tcase:
        runner.run([], smap, tcase)
        assert runner.result == runner.COMPLETE
        assert "inc_file.bin" in tcase._existing_paths
        assert pathjoin("nested", "nested_inc.bin") in tcase._existing_paths
        assert pathjoin("test", "inc_file3.txt") in tcase._existing_paths

def test_idle_check_01(mocker):
    """test simple _IdleChecker"""
    fake_time = mocker.patch("grizzly.common.runner.time", autospec=True)
    ichk = _IdleChecker(mocker.Mock(), 100, 10, poll_delay=1)
    assert ichk._threshold == 100
    assert ichk._init_delay == 10
    assert ichk._poll_delay == 1
    assert ichk._next_poll is None
    fake_time.return_value = 0
    ichk.schedule_poll(initial=True)
    assert ichk._next_poll == 10
    ichk.schedule_poll()
    assert ichk._next_poll == 1

def test_idle_check_02(mocker):
    """test _IdleChecker.is_idle()"""
    fake_time = mocker.patch("grizzly.common.runner.time", autospec=True)
    callbk = mocker.Mock()
    callbk.return_value = False
    #check_cb, delay, duration, threshold
    ichk = _IdleChecker(callbk, 100, 10, poll_delay=1)
    fake_time.return_value = 0
    ichk.schedule_poll()
    # early check
    assert not ichk.is_idle()
    assert callbk.call_count == 0
    # not idle
    fake_time.return_value = 10
    assert not ichk.is_idle()
    assert ichk._next_poll == 11
    assert callbk.call_count == 1
    # idle
    callbk.return_value = True
    fake_time.return_value = ichk._next_poll
    assert ichk.is_idle()
    assert callbk.call_count == 2



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from logging import getLogger
from os.path import dirname, join as pathjoin
from tempfile import mkdtemp
from time import sleep

from FTB.Signatures.CrashInfo import CrashSignature
from sapphire import Sapphire, ServerMap

from ..common.reporter import FilesystemReporter, FuzzManagerReporter, Report
from ..common.runner import Runner
from ..common.status import Status
from ..common.storage import TestCase, TestCaseLoadFailure, TestFile
from ..common.utils import grz_tmp
from ..main import configure_logging
from ..target import load as load_target, TargetLaunchError, TargetLaunchTimeout

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger("replay")


class ReplayManager(object):
    HARNESS_FILE = pathjoin(dirname(__file__), "..", "common", "harness.html")

    __slots__ = ("ignore", "server", "status", "target", "testcase", "_any_crash",
                 "_harness", "_reports_expected", "_reports_other", "_runner",
                 "_signature")

    def __init__(self, ignore, server, target, testcase, any_crash=False, signature=None, use_harness=True):
        self.ignore = ignore
        self.server = server
        self.status = None
        self.target = target
        self.testcase = testcase
        self._any_crash = any_crash
        self._harness = None
        self._reports_expected = dict()
        self._reports_other = dict()
        self._runner = Runner(self.server, self.target)
        self._signature = signature

        if use_harness:
            with open(self.HARNESS_FILE, "rb") as in_fp:
                self._harness = in_fp.read()

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    def cleanup(self):
        """Remove temporary files from disk.

        Args:
            None

        Returns:
            None
        """
        for report in self._reports_expected.values():
            report.cleanup()
        self._reports_expected.clear()
        for report in self._reports_other.values():
            report.cleanup()
        self._reports_other.clear()
        if self.status is not None:
            self.status.cleanup()

    @property
    def other_reports(self):
        """Reports from results that do not match:
            - the given signature
            - the initial result (if any-crash is not specified)

        Args:
            None

        Returns:
            iterable: Reports.
        """
        return self._reports_other.values()

    @property
    def reports(self):
        """Reports from results.

        Args:
            None

        Returns:
            iterable: Reports.
        """
        return self._reports_expected.values()

    @staticmethod
    def report_to_filesystem(path, reports, other_reports=None, test=None):
        """Use FilesystemReporter to write reports and testcase to disk in a
        known location.

        Args:
            path (str): Location to write data.
            reports (iterable): Reports to output.
            other_reports (iterable): Reports to output.
            test (TestCase): Testcase to output.

        Returns:
            None
        """
        assert test is None or isinstance(test, TestCase)
        tests = [test] if test else tuple()
        if reports:
            reporter = FilesystemReporter(
                report_path=pathjoin(path, "reports"),
                major_bucket=False)
            for report in reports:
                reporter.submit(tests, report=report)
        if other_reports:
            reporter = FilesystemReporter(
                report_path=pathjoin(path, "other_reports"),
                major_bucket=False)
            for report in other_reports:
                reporter.submit(tests, report=report)

    def run(self, repeat=1, min_results=1):
        """Run testcase replay.

        Args:
            repeat (int): Maximum number of times to run the test case.
            min_results (int): Minimum number of results needed before run can
                               be considered successful.

        Returns:
            bool: True if results were reproduced otherwise False.
        """
        assert repeat > 0
        assert min_results > 0
        assert min_results <= repeat
        self.status = Status.start()

        server_map = ServerMap()
        if self._harness is not None:
            def _dyn_close():  # pragma: no cover
                if self.target.monitor.is_healthy():
                    # delay to help catch window close/shutdown related crashes
                    sleep(0.1)
                    self.target.close()
                return b"<h1>Close Browser</h1>"
            server_map.set_dynamic_response("grz_close_browser", _dyn_close, mime_type="text/html")
            server_map.set_dynamic_response("grz_harness", lambda: self._harness, mime_type="text/html")
            server_map.set_redirect("grz_next_test", self.testcase.landing_page, required=True)
        server_map.set_redirect("grz_current_test", self.testcase.landing_page, required=False)

        success = False
        for _ in range(repeat):
            self.status.iteration += 1
            if self.target.closed:
                LOG.info("Launching target...")
                if self._harness is None:
                    location = self._runner.location(
                        "/grz_current_test",
                        self.server.port)
                else:
                    location = self._runner.location(
                        "/grz_harness",
                        self.server.port,
                        close_after=self.target.rl_reset,
                        forced_close=self.target.forced_close)
                try:
                    self._runner.launch(location, env_mod=self.testcase.env_vars)
                except TargetLaunchError:
                    LOG.error("Target launch error. Check browser logs for details.")
                    log_path = mkdtemp(prefix="logs_", dir=grz_tmp("logs"))
                    self.target.save_logs(log_path)
                    self._reports_other["STARTUP"] = Report.from_path(log_path)
                    raise
            self.target.step()
            LOG.info("Performing replay (%d/%d)...", self.status.iteration, repeat)
            # run test case
            self._runner.run(
                self.ignore,
                server_map,
                self.testcase,
                wait_for_callback=self._harness is None)
            # process results
            if self._runner.result == self._runner.FAILED:
                log_path = mkdtemp(prefix="logs_", dir=grz_tmp("logs"))
                self.target.save_logs(log_path)
                report = Report.from_path(log_path)
                # check signatures
                crash_info = report.crash_info(self.target.binary)
                short_sig = crash_info.createShortSignature()
                if not self._any_crash and self._signature is None and short_sig != "No crash detected":
                    # signature has not been specified use the first one created
                    self._signature = report.crash_signature(crash_info)
                if short_sig == "No crash detected":
                    # TODO: verify report.major == "NO_STACK" otherwise FM failed to parse the logs
                    # TODO: change this to support hangs/timeouts, etc
                    LOG.info("Result: No crash detected")
                    crash_hash = None
                elif self._any_crash or self._signature.matches(crash_info):
                    self.status.results += 1
                    LOG.info("Result: %s (%s:%s)",
                             short_sig, report.major[:8], report.minor[:8])
                    crash_hash = report.crash_hash(crash_info)
                    if crash_hash not in self._reports_expected:
                        LOG.debug("now tracking %s", crash_hash)
                        self._reports_expected[crash_hash] = report
                        report = None  # don't remove report
                    assert self._any_crash or len(self._reports_expected) == 1
                else:
                    LOG.info("Result: Different signature: %s (%s:%s)",
                             short_sig, report.major[:8], report.minor[:8])
                    self.status.ignored += 1
                    crash_hash = report.crash_hash(crash_info)
                    if crash_hash not in self._reports_other:
                        LOG.debug("now tracking %s", crash_hash)
                        self._reports_other[crash_hash] = report
                        report = None  # don't remove report
                # purge untracked report
                if report is not None:
                    if crash_hash is not None:
                        LOG.debug("already tracking %s", crash_hash)
                    report.cleanup()
                    report = None
            elif self._runner.result == self._runner.IGNORED:
                self.status.ignored += 1
                LOG.info("Result: Ignored (%d)", self.status.ignored)
            elif self._runner.result == self._runner.ERROR:
                LOG.error("ERROR: Replay malfunction, test case was not served")
                break

            # check status and exit early if possible
            if repeat - self.status.iteration + self.status.results < min_results:
                if self.status.iteration < repeat:
                    LOG.debug("skipping remaining attempts")
                # failed to reproduce issue
                LOG.debug("results (%d) < expected (%s) after %d attempts",
                          self.status.results, min_results, self.status.iteration)
                break
            if self.status.results >= min_results:
                assert self.status.results == min_results
                success = True
                LOG.debug("results == expected (%s) after %d attempts",
                          min_results, self.status.iteration)
                break

            # warn about large browser logs
            #self.status.log_size = self.target.log_size()
            #if self.status.log_size > self.TARGET_LOG_SIZE_WARN:
            #    LOG.warning("Large browser logs: %dMBs", (self.status.log_size / 0x100000))

            # trigger relaunch by closing the browser if needed
            self.target.check_relaunch()

        if success:
            LOG.info("Result successfully reproduced")
        else:
            LOG.info("Failed to reproduce results")
        self.target.close()
        return success

    @classmethod
    def main(cls, args):
        configure_logging(args.log_level)
        if args.fuzzmanager:
            FuzzManagerReporter.sanity_check(args.binary)

        LOG.info("Starting Grizzly Replay")

        if args.ignore:
            LOG.info("Ignoring: %s", ", ".join(args.ignore))
        if args.xvfb:
            LOG.info("Running with Xvfb")
        if args.valgrind:
            LOG.info("Running with Valgrind. This will be SLOW!")
        if args.rr:
            LOG.info("Running with RR")

        if args.sig:
            signature = CrashSignature.fromFile(args.sig)
        else:
            signature = None

        try:
            LOG.debug("loading the TestCase")
            testcase = TestCase.load_path(args.input, prefs=args.prefs is None)
            # prioritize specified prefs.js file over included file
            if args.prefs is not None:
                prefs = args.prefs
                testcase.add_meta(TestFile.from_file(args.prefs, "prefs.js"))
                LOG.debug("using specified prefs.js")
            elif testcase.contains("prefs.js"):
                prefs = pathjoin(args.input, "prefs.js")
                LOG.debug("using included prefs.js")
            else:
                prefs = None
        except TestCaseLoadFailure as exc:
            LOG.error("Error: %s", str(exc))
            return 1

        replay = None
        target = None
        try:
            relaunch = min(args.relaunch, args.repeat)
            LOG.debug("initializing the Target")
            target = load_target(args.platform)(
                args.binary,
                args.extension,
                args.launch_timeout,
                args.log_limit,
                args.memory,
                prefs,
                relaunch,
                rr=args.rr,
                valgrind=args.valgrind,
                xvfb=args.xvfb)
            if testcase.env_vars.get("GRZ_FORCED_CLOSE") == "0":
                LOG.debug("setting target.forced_close=False")
                target.forced_close = False

            LOG.debug("starting sapphire server")
            # launch HTTP server used to serve test cases
            with Sapphire(auto_close=1, timeout=args.timeout) as server:
                target.reverse(server.port, server.port)
                if args.no_harness:
                    LOG.debug("--no-harness specified relaunch set to 1")
                    args.relaunch = 1
                args.repeat = max(args.min_crashes, args.repeat)
                LOG.info("Repeat: %d, Minimum crashes: %d, Relaunch %d",
                         args.repeat, args.min_crashes, relaunch)
                replay = ReplayManager(
                    args.ignore,
                    server,
                    target,
                    testcase,
                    any_crash=args.any_crash,
                    signature=signature,
                    use_harness=not args.no_harness)
                success = replay.run(repeat=args.repeat, min_results=args.min_crashes)
            if args.logs:
                replay.report_to_filesystem(
                    args.logs,
                    replay.reports,
                    replay.other_reports,
                    replay.testcase if args.include_test else None)
            # TODO: add fuzzmanager reporting
            return 0 if success else 1

        except KeyboardInterrupt:
            return 1

        except (TargetLaunchError, TargetLaunchTimeout):
            if args.logs:
                replay.report_to_filesystem(
                    args.logs,
                    replay.reports,
                    replay.other_reports)
            return 1

        finally:
            LOG.warning("Shutting down...")
            if replay is not None:
                replay.cleanup()
            if target is not None:
                target.cleanup()
            if testcase is not None:
                testcase.cleanup()
            LOG.info("Done.")



#!/usr/bin/env python
# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Manage Grizzly status reports."""

import argparse
from datetime import timedelta
from functools import partial
import logging
import os
import re
import sys
import time

import psutil

from .status import ReducerStats, Status

__all__ = ("StatusReporter",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

class StatusReporter(object):
    """Read and merge Grizzly status reports, including tracebacks if found.
    Output is a single textual report, e.g. for submission to EC2SpotManager.
    """
    CPU_POLL_INTERVAL = 1
    DISPLAY_LIMIT_LOG = 10  # don't include log results unless size exceeds 10MBs
    EXP_LIMIT = 600  # expiration limit, ignore older reports
    READ_BUF_SIZE = 0x10000  # 64KB
    SUMMARY_LIMIT = 4095  # summary output must be no more than 4KB

    def __init__(self, reports, reducer=False, tracebacks=None):
        self._reducer = reducer
        self.reports = reports
        self.tracebacks = tracebacks

    def dump_specific(self, filename):
        """Write out merged reports.

        Args:
            filename (str): Path where output should be written.

        Returns:
            None
        """
        with open(filename, "w") as ofp:
            ofp.write(self._specific())

    def dump_summary(self, filename, runtime=False, sysinfo=True, timestamp=True):
        """Write out summary merged reports.

        Args:
            filename (str): Path where output should be written.
            runtime (bool): Include total runtime in output
            sysinfo (bool): Include system info (CPU, disk, RAM... etc) in output
            timestamp (bool): Include time stamp in output

        Returns:
            None
        """
        with open(filename, "w") as ofp:
            ofp.write(self._summary(runtime=runtime, sysinfo=sysinfo, timestamp=timestamp))

    @classmethod
    def load(cls, reducer=False, tb_path=None):
        """Read Grizzly status reports and create a StatusReporter object

        Args:
            tb_path (str): Directory to scan for files containing Python tracebacks

        Returns:
            StatusReporter: Contains status reports and traceback reports that were found
        """
        if tb_path is not None and not os.path.isdir(tb_path):
            raise IOError("%r is not a directory" % (tb_path,))
        tracebacks = None if tb_path is None else cls._tracebacks(tb_path)
        return cls(list(Status.loadall()), reducer=reducer, tracebacks=tracebacks)

    def print_specific(self):
        print(self._specific())

    def print_summary(self, runtime=True, sysinfo=False, timestamp=False):
        print(self._summary(runtime=runtime, sysinfo=sysinfo, timestamp=timestamp))

    @staticmethod
    def _scan(path, fname_pattern):
        abs_path = os.path.abspath(path)
        for fname in os.listdir(abs_path):
            if fname_pattern.match(fname) is None:
                continue
            full_path = os.path.join(abs_path, fname)
            if not os.path.isfile(full_path):
                continue
            if os.path.getsize(full_path) > 0:
                yield full_path

    def _specific(self):
        """Merged and generate formatted output of status reports.

        Args:
            None

        Returns:
            str: A formatted report
        """
        if not self.reports:
            return "No status reports available"
        exp = int(time.time()) - self.EXP_LIMIT
        self.reports.sort(key=lambda x: x.duration, reverse=True)
        self.reports.sort(key=lambda x: x.timestamp < exp)
        txt = list()
        for num, report in enumerate(self.reports, start=1):
            txt.append("#%02d" % (num,))
            if report.timestamp < exp:
                txt.append(" (EXPIRED)\n")
                continue
            txt.append(" Runtime %s\n" % str(timedelta(seconds=int(report.duration))))
            txt.append(" * Iterations: %03d" % report.iteration)
            txt.append(" - Rate: %0.2f" % report.rate)
            if not self._reducer:
                txt.append(" - Ignored: %02d" % report.ignored)
                txt.append(" - Results: %d" % report.results)
            txt.append("\n")
        return "".join(txt)

    def _summary(self, runtime=True, sysinfo=False, timestamp=False):
        """Merge and generate a summary of status reports.

        Args:
            filename (str): Path where output should be written.
            runtime (bool): Include total runtime in output
            sysinfo (bool): Include system info (CPU, disk, RAM... etc) in output
            timestamp (bool): Include time stamp in output

        Returns:
            str: A summary of merged reports
        """
        txt = list()
        if self._reducer:
            # reducer stats
            with ReducerStats() as stats:
                r_error = stats.error
                r_failed = stats.failed
                r_passed = stats.passed
            txt.append("======== Stats ========\n")
            # Reduced successfully
            txt.append("   Reduced : %d\n" % (r_passed,))
            # Failed to reproduce
            txt.append("  No Repro : %d\n" % (r_failed,))
            # Error during reduction
            txt.append("    Errors : %d\n" % (r_error,))
            txt.append("======= Active ========\n")
        # Job specific status
        exp = int(time.time()) - self.EXP_LIMIT
        reports = tuple(x for x in self.reports if x.timestamp > exp)
        if reports:
            # calculate totals
            iterations = tuple(x.iteration for x in reports)
            log_sizes = tuple(x.log_size for x in reports)
            rates = tuple(x.rate for x in reports)
            results = tuple(x.results for x in reports)
            count = len(reports)
            total_ignored = sum(x.ignored for x in reports)
            total_iters = sum(iterations)
            # Iterations
            txt.append("Iterations : %d" % (total_iters,))
            if count > 1:
                txt.append(" (%s, %s)" % (max(iterations), min(iterations)))
            txt.append("\n")
            # Rate
            txt.append("      Rate : %d @ %0.2f" % (count, sum(rates)))
            if count > 1:
                txt.append(" (%0.2f, %0.2f)" % (max(rates), min(rates)))
            txt.append("\n")
            # Results / Signature mismatch
            if self._reducer:
                txt.append("  Mismatch : %d" % (sum(results),))
            else:
                txt.append("   Results : %d" % (sum(results),))
            if total_ignored:
                ignore_pct = (total_ignored / float(total_iters)) * 100
                txt.append(" (%d ignored @ %0.2f%%)" % (total_ignored, ignore_pct))
            # Runtime
            if runtime:
                txt.append("\n")
                if self._reducer:
                    durations = tuple(x.duration for x in reports)
                    if count > 1:
                        max_duration = str(timedelta(seconds=int(max(durations))))
                        min_duration = str(timedelta(seconds=int(min(durations))))
                        txt.append("   Runtime : (%s, %s)" % (max_duration, min_duration))
                    else:
                        txt.append("   Runtime : %s" % (str(timedelta(seconds=int(durations[0]))),))
                else:
                    total_runtime = sum(x.duration for x in reports)
                    txt.append("   Runtime : %s" % (str(timedelta(seconds=int(total_runtime))),))
            # Log size
            log_usage = sum(log_sizes) / 1048576.0
            if log_usage > self.DISPLAY_LIMIT_LOG:
                txt.append("\n")
                txt.append("      Logs : %0.1fMB" % (log_usage,))
                if count > 1:
                    txt.append(" (%0.2fMB, %0.2fMB)" % (
                        max(log_sizes) / 1048576.0,
                        min(log_sizes) / 1048576.0))
        else:
            txt.append("No status reports available")
        if sysinfo:
            txt.append("\n")
            txt.append(self._sys_info())
        if timestamp:
            txt.append("\n")
            txt.append(time.strftime(" Timestamp : %Y/%m/%d %X %z", time.gmtime()))
        msg = "".join(txt)
        if self.tracebacks:
            txt = self._merge_tracebacks(self.tracebacks, self.SUMMARY_LIMIT - len(msg))
            msg = "".join((msg, txt))
        return msg

    @staticmethod
    def _merge_tracebacks(tracebacks, size_limit):
        """Merge traceback without exceeding size_limit.

        Args:
            tracebacks (iterable): TracebackReport to merge.
            size_limit (int): Maximum size in bytes of output.

        Returns:
            str: merged tracebacks.
        """
        txt = list()
        txt.append("\n\nWARNING Tracebacks (%d) detected!" % (len(tracebacks),))
        tb_size = len(txt[-1])
        for tbr in tracebacks:
            tb_size += len(tbr) + 1
            if tb_size > size_limit:
                break
            txt.append(str(tbr))
        return "\n".join(txt)

    @staticmethod
    def _sys_info():
        """Collect and format system information.

        Args:
            None

        Returns:
            str: System information formatted to match output from _summary()
        """
        txt = list()
        txt.append("CPU & Load : %d @ %0.1f%%" % (
            psutil.cpu_count(),
            psutil.cpu_percent(interval=StatusReporter.CPU_POLL_INTERVAL)))
        try:
            txt.append(" %s\n" % (str(os.getloadavg()),))
        except AttributeError:
            txt.append("\n")
        mem_usage = psutil.virtual_memory()
        txt.append("    Memory : ")
        if mem_usage.available < 1073741824:  # < 1GB
            txt.append("%dMB" % (mem_usage.available / 1048576,))
        else:
            txt.append("%0.1fGB" % (mem_usage.available / 1073741824.0,))
        txt.append(" of %0.1fGB free\n" % (mem_usage.total / 1073741824.0,))
        disk_usage = psutil.disk_usage("/")
        txt.append("      Disk : ")
        if disk_usage.free < 1073741824:  # < 1GB
            txt.append("%dMB" % (disk_usage.free / 1048576,))
        else:
            txt.append("%0.1fGB" % (disk_usage.free / 1073741824.0,))
        txt.append(" of %0.1fGB free" % (disk_usage.total / 1073741824.0,))
        return "".join(txt)

    @staticmethod
    def _tracebacks(path, ignore_kbi=True, max_preceeding=5):
        """Search screen logs for tracebacks.

        Args:
            path (str): Directory containing log files.
            ignore_kbi (bool): Do not include KeyboardInterupts in results
            max_preceeding (int): Maximum number of lines preceding traceback to include.

        Returns:
            list: A list of TracebackReports
        """
        tracebacks = list()
        for screen_log in StatusReporter._scan(path, re.compile(r"screenlog\.\d+")):
            tbr = TracebackReport.from_file(screen_log, max_preceeding=max_preceeding)
            if tbr is None:
                continue
            if ignore_kbi and tbr.is_kbi:
                continue
            tracebacks.append(tbr)
        return tracebacks


class TracebackReport(object):
    """Read Python tracebacks from log files and store it in a manner that is helpful
    when generating reports.
    """
    MAX_LINES = 16  # should be no less than 6
    READ_LIMIT = 0x10000  # 64KB

    def __init__(self, file_name, lines, is_kbi=False, prev_lines=None):
        assert isinstance(lines, list)
        self.file_name = file_name
        self.lines = lines
        self.prev_lines = list() if prev_lines is None else prev_lines
        self.is_kbi = is_kbi

    @classmethod
    def from_file(cls, input_log, max_preceeding=5):
        """Create TracebackReport from a text file containing a Python traceback.
        Only the first traceback in the file will be parsed.

        Args:
            input_log (str): File to parse.
            max_preceeding (int): Number of lines to collect leading up to the traceback.

        Returns:
            TracebackReport: Contains data from input_log.
        """
        token = b"Traceback (most recent call last):"
        assert len(token) < cls.READ_LIMIT
        try:
            with open(input_log, "rb") as in_fp:
                for chunk in iter(partial(in_fp.read, cls.READ_LIMIT), b""):
                    idx = chunk.find(token)
                    if idx > -1:
                        # calculate offset of data in the file
                        pos = in_fp.tell() - len(chunk) + idx
                        break
                    if len(chunk) == cls.READ_LIMIT:
                        # seek back to avoid missing beginning of token
                        in_fp.seek(len(token) * -1, os.SEEK_CUR)
                else:
                    # no traceback here, move along
                    return None
                # seek back 2KB to collect preceding lines
                in_fp.seek(max(pos - 2048, 0))
                data = in_fp.read(cls.READ_LIMIT)
        except IOError:  # pragma: no cover
            # in case the file goes away
            return None

        data = data.decode("ascii", errors="ignore").splitlines()
        token = token.decode()
        is_kbi = False
        tb_start = None
        tb_end = None
        line_count = len(data)
        for line_num, log_line in enumerate(data):
            if tb_start is None and token in log_line:
                tb_start = line_num
                continue
            if tb_start is not None:
                log_line = log_line.strip()
                if not log_line:
                    # stop at first empty line
                    tb_end = min(line_num, line_count)
                    break
                if re.match(r"^\w+(\.\w+)*\:\s|^\w+(Interrupt|Error)$", log_line):
                    is_kbi = log_line.startswith("KeyboardInterrupt")
                    # stop after error message
                    tb_end = min(line_num + 1, line_count)
                    break
        assert tb_start is not None
        if max_preceeding > 0:
            prev_start = max(tb_start - max_preceeding, 0)
            prev_lines = data[prev_start:tb_start]
        else:
            prev_lines = None
        if tb_end is None:
            # limit if the end is not identified (failsafe)
            tb_end = max(line_count, cls.MAX_LINES)
        if tb_end - tb_start > cls.MAX_LINES:
            # add first entry
            lines = data[tb_start:tb_start + 3]
            lines += ["<--- TRACEBACK TRIMMED--->"]
            # add end entries
            lines += data[tb_end - (cls.MAX_LINES - 3):tb_end]
        else:
            lines = data[tb_start:tb_end]
        return cls(input_log, lines, is_kbi=is_kbi, prev_lines=prev_lines)

    def __len__(self):
        return len(str(self))

    def __str__(self):
        return "\n".join(["Log: %r" % self.file_name] + self.prev_lines + self.lines)


def main(args=None):
    """Merge Grizzly status files into a single report (main entrypoint).

    Args:
        args (list/None): Argument list to parse instead of sys.argv (for testing).

    Returns:
        None
    """
    log_level = logging.INFO
    log_fmt = "[%(asctime)s] %(message)s"
    if bool(os.getenv("DEBUG")):  # pragma: no cover
        log_level = logging.DEBUG
        log_fmt = "%(levelname).1s %(name)s [%(asctime)s] %(message)s"
    logging.basicConfig(format=log_fmt, datefmt="%Y-%m-%d %H:%M:%S", level=log_level)

    modes = ("reduce-status", "status")
    parser = argparse.ArgumentParser(description="Grizzly status report generator")
    parser.add_argument(
        "--dump",
        help="File to write report to")
    parser.add_argument(
        "--mode", default="status",
        help="Status loading mode. Available modes: %s (default: 'status')" % (", ".join(modes),))
    parser.add_argument(
        "--system-report", action="store_true",
        help="Output summary and system information")
    parser.add_argument(
        "--tracebacks",
        help="Scan path for Python tracebacks found in screenlog.# files")
    args = parser.parse_args(args)

    if args.mode not in modes:
        parser.error("Invalid mode %r" % args.mode)
    reducer_mode = args.mode == "reduce-status"
    reporter = StatusReporter.load(tb_path=args.tracebacks, reducer=reducer_mode)
    if args.dump:
        reporter.dump_summary(args.dump, runtime=reducer_mode)
        return 0
    if not reporter.reports:
        print("No status reports to display")
        return 0
    print("Grizzly Status Report")
    print("---------------------")
    print("Status report frequency: %ds\n" % (Status.REPORT_FREQ,))
    reporter.print_specific()
    print("Summary")
    print("-------")
    reporter.print_summary(sysinfo=args.system_report)
    return 0


if __name__ == "__main__":
    sys.exit(main())



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import basicConfig, DEBUG, getLogger

from sapphire import Sapphire

from .adapters import get as get_adapter
from .common import FilesystemReporter, FuzzManagerReporter, IOManager, S3FuzzManagerReporter
from .session import Session
from .target import load as load_target, TargetLaunchError, TargetLaunchTimeout


__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]


log = getLogger("grizzly")  # pylint: disable=invalid-name

def configure_logging(log_level):
    if log_level == DEBUG:
        log_fmt = "%(levelname).1s %(name)s [%(asctime)s] %(message)s"
    else:
        log_fmt = "[%(asctime)s] %(message)s"
    basicConfig(format=log_fmt, datefmt="%Y-%m-%d %H:%M:%S", level=log_level)

def main(args):
    configure_logging(args.log_level)
    log.info("Starting Grizzly")
    if args.fuzzmanager:
        FuzzManagerReporter.sanity_check(args.binary)
    elif args.s3_fuzzmanager:
        S3FuzzManagerReporter.sanity_check(args.binary)

    if args.ignore:
        log.info("Ignoring: %s", ", ".join(args.ignore))
    if args.xvfb:
        log.info("Running with Xvfb")
    if args.valgrind:
        log.info("Running with Valgrind. This will be SLOW!")
    if args.rr:
        log.info("Running with RR")

    adapter = None
    iomanager = None
    session = None
    target = None
    try:
        log.debug("initializing the IOManager")
        # TODO: move this into Session
        iomanager = IOManager(report_size=(max(args.cache, 0) + 1))

        log.debug("initializing Adapter %r", args.adapter)
        adapter = get_adapter(args.adapter)()

        if adapter.TEST_DURATION >= args.timeout:
            raise RuntimeError("Test duration (%ds) should be less than browser timeout (%ds)" % (
                adapter.TEST_DURATION, args.timeout))

        if adapter.RELAUNCH > 0:
            log.debug("relaunch (%d) set in Adapter", adapter.RELAUNCH)
            relaunch = adapter.RELAUNCH
        else:
            relaunch = args.relaunch

        log.debug("initializing the Target")
        target = load_target(args.platform)(
            args.binary,
            args.extension,
            args.launch_timeout,
            args.log_limit,
            args.memory,
            args.prefs,
            relaunch,
            rr=args.rr,
            valgrind=args.valgrind,
            xvfb=args.xvfb)
        adapter.monitor = target.monitor

        if args.coverage and relaunch == 1 and target.forced_close:
            # this is a workaround to avoid not dumping coverage
            # GRZ_FORCED_CLOSE=0 is also an option but the browser MUST
            # close itself.
            raise RuntimeError("Coverage must be run with --relaunch > 1")

        log.debug("calling adapter setup()")
        adapter.setup(args.input, iomanager.server_map)
        log.debug("configuring harness")
        iomanager.harness = adapter.get_harness()

        log.debug("initializing the Reporter")
        if args.fuzzmanager:
            log.info("Results will be reported via FuzzManager")
            reporter = FuzzManagerReporter(args.binary, tool=args.tool)
        elif args.s3_fuzzmanager:
            log.info("Results will be reported via FuzzManager w/ large attachments in S3")
            reporter = S3FuzzManagerReporter(args.binary, tool=args.tool)
        else:
            reporter = FilesystemReporter()
            log.info("Results will be stored in %r", reporter.report_path)

        # set 'auto_close=1' so the client error pages (code 4XX) will
        # call 'window.close()' after a second.
        # launch http server used to serve test cases
        log.debug("starting Sapphire server")
        with Sapphire(auto_close=1, timeout=args.timeout) as server:
            target.reverse(server.port, server.port)
            log.debug("initializing the Session")
            session = Session(
                adapter,
                iomanager,
                reporter,
                server,
                target,
                coverage=args.coverage)
            if args.log_level == DEBUG or args.verbose:
                display_mode = Session.DISPLAY_VERBOSE
            else:
                display_mode = Session.DISPLAY_NORMAL
            session.run(args.ignore, display_mode=display_mode)

    except KeyboardInterrupt:
        log.info("Ctrl+C detected.")
        return Session.EXIT_ABORT

    except (TargetLaunchError, TargetLaunchTimeout):
        return Session.EXIT_LAUNCH_FAILURE

    finally:
        log.warning("Shutting down...")
        if session is not None:
            log.debug("calling session.close()")
            session.close()
        if target is not None:
            log.debug("calling target.cleanup()")
            target.cleanup()
        if adapter is not None:
            log.debug("calling adapter.cleanup()")
            adapter.cleanup()
        if iomanager is not None:
            log.debug("calling iomanager.cleanup()")
            iomanager.cleanup()
        log.info("Done.")

    return Session.EXIT_SUCCESS



# coding=utf-8
"""
Sapphire HTTP server
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .core import Sapphire
from .server_map import ServerMap
from .status_codes import SERVED_ALL, SERVED_NONE, SERVED_REQUEST, SERVED_TIMEOUT

__all__ = ("Sapphire", "SERVED_ALL", "SERVED_NONE", "SERVED_REQUEST", "SERVED_TIMEOUT", "ServerMap")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import abc
import logging
import math
import os
import re

try:
    import cssbeautifier
    HAVE_CSSBEAUTIFIER = True
except ImportError:
    HAVE_CSSBEAUTIFIER = False
try:
    import jsbeautifier
    HAVE_JSBEAUTIFIER = True
except ImportError:
    HAVE_JSBEAUTIFIER = False
import lithium

from . import testcase_contents


LOG = logging.getLogger("grizzly.reduce.strategies")


class ReduceStage(metaclass=abc.ABCMeta):
    strategy_type = None
    testcase_type = None

    def __init__(self, job, run_state, reducer):
        if getattr(self, "USES_ANALYSIS_MODE", False):
            self._analysis_mode = job.analysis_mode
        else:
            self._analysis_mode = None

        if getattr(self, "ALTERS_JOB_TESTCASE", False):
            self.job_testcase = job.testcase_proxy(run_state)

        if getattr(self, "ALTERS_JOB_TIMEOUTS", False):
            self.job_timeouts = job.timeouts_proxy()

        self.lithium = reducer

    def read_testcase(self, testcase_path):
        if self._analysis_mode is not None:
            self.lithium.strategy = self.strategy_type(self._analysis_mode)  # pylint: disable=not-callable
        else:
            self.lithium.strategy = self.strategy_type()  # pylint: disable=not-callable
        self.lithium.testcase = self.testcase_type()  # pylint: disable=not-callable
        LOG.info("Reducing %s with %s on %ss",
                 testcase_path, self.strategy_type.name, self.testcase_type.atom)
        self.lithium.testcase.readTestcase(testcase_path)

    def should_skip(self):  # pylint: disable=no-self-use
        return False

    def on_success(self):
        LOG.info("%s succeeded", type(self).__name__)

    def on_failure(self):  # pylint: disable=no-self-use
        raise StopIteration()


class MinimizeLines(ReduceStage):
    name = "line"
    strategy_type = lithium.Minimize
    testcase_type = lithium.TestcaseLine


class _AnalyzeReliability(lithium.Strategy):
    name = "reliability-analysis"
    ITERATIONS = 11  # number of iterations to analyze
    MIN_CRASHES = 2  # --min-crashes value when analysis is used
    TARGET_PROBABILITY = 0.95  # probability that successful reduction will observe the crash
    # to see the worst case, run the `self.repeat` calculation below using
    # crashes_percent = 1.0/ITERATIONS

    def __init__(self, analysis_mode_cb):
        super(_AnalyzeReliability, self).__init__()
        self.analysis_mode_cb = analysis_mode_cb

    def main(self, testcase, interesting, _temp_filename):

        assert self.ITERATIONS > 0

        harness_crashes = 0
        non_harness_crashes = 0

        # Reset parameters.
        # Use repeat=1 & relaunch=ITERATIONS because this is closer to how we will run post-analysis.
        # We're only using repeat=1 instead of repeat=ITERATIONS so we can get feedback on every
        #   call to interesting.
        with self.analysis_mode_cb(min_crashes=1, relaunch=self.ITERATIONS, repeat=1) as iter_params:

            iter_params.commit()

            if not iter_params.force_no_harness:
                LOG.info("Running for %d iterations to assess reliability using harness.", self.ITERATIONS)
                for _ in range(self.ITERATIONS):
                    result = interesting(testcase, writeIt=False)  # pylint: disable=invalid-name
                    LOG.info("Lithium result: %s", "interesting." if result else "not interesting.")
                    if result:
                        harness_crashes += 1
                LOG.info("Testcase was interesting %0.1f%% of %d attempts using harness for iteration.",
                         100.0 * harness_crashes / self.ITERATIONS, self.ITERATIONS)

                # close target so new parameters take effect
                iter_params.commit()

            if harness_crashes != self.ITERATIONS:
                # try without harness
                iter_params.no_harness = True

                LOG.info("Running for %d iterations to assess reliability without harness.",
                         self.ITERATIONS)
                for _ in range(self.ITERATIONS):
                    result = interesting(testcase, writeIt=False)  # pylint: disable=invalid-name
                    LOG.info("Lithium result: %s", "interesting." if result else "not interesting.")
                    if result:
                        non_harness_crashes += 1
                LOG.info("Testcase was interesting %0.1f%% of %d attempts without harness.",
                         100.0 * non_harness_crashes / self.ITERATIONS, self.ITERATIONS)

                # close target so new parameters take effect
                iter_params.commit()

        if harness_crashes == 0 and non_harness_crashes == 0:
            return 1  # no crashes ever?

        # should we use the harness? go with whichever crashed more
        iter_params.no_harness = non_harness_crashes > harness_crashes

        # this is max 99% to avoid domain errors in the calculation below
        crashes_percent = min(1.0 * max(non_harness_crashes, harness_crashes) / self.ITERATIONS, 0.99)

        # adjust repeat/min-crashes depending on how reliable the testcase was
        iter_params.min_crashes = self.MIN_CRASHES
        iter_params.repeat = \
            int(math.ceil(math.log(1 - self.TARGET_PROBABILITY, 1 - crashes_percent)) * self.MIN_CRASHES)
        iter_params.relaunch = iter_params.repeat  # actual value used will respect relaunch requested by user

        LOG.info("Analysis results:")
        if harness_crashes == self.ITERATIONS:
            LOG.info("* testcase was perfectly reliable with the harness (--no-harness not assessed)")
        elif harness_crashes == non_harness_crashes:
            LOG.info("* testcase was equally reliable with/without the harness")
        elif iter_params.force_no_harness:
            LOG.info("* --no-harness was already set")
        else:
            LOG.info("* testcase was %s reliable with the harness",
                     "less" if iter_params.no_harness else "more")
        LOG.info("* adjusted parameters: --min-crashes=%d --repeat=%d --relaunch=%d",
                 iter_params.min_crashes, iter_params.repeat,
                 iter_params.relaunch)

        return 0


class AnalyzeTestcase(ReduceStage):
    name = "analyze"
    USES_ANALYSIS_MODE = True
    strategy_type = _AnalyzeReliability
    testcase_type = lithium.TestcaseLine

    def on_success(self):
        super(AnalyzeTestcase, self).on_success()
        # only run this strategy once, not once per reducible file in the testcase
        raise StopIteration()


class MinimizeCacheIterHarness(MinimizeLines):
    name = "minimize-cache"
    ALTERS_JOB_TESTCASE = True
    ALTERS_JOB_TIMEOUTS = True

    # job attributes:
    # write: files_to_reduce (clear, append), tcroot, testcase, landing_page
    # read: cache_iter_harness_created

    def should_skip(self):
        return not self.job_testcase.cache_iter_harness_created

    def read_testcase(self, testcase_path):
        if self.should_skip():
            return

        super(MinimizeCacheIterHarness, self).read_testcase(testcase_path)

        # we are running multiple testcases in a single "iteration", so we need to
        #   fix the timeout values

        # start polling for idle after n-1 testcases have finished
        self.job_timeouts.idle += \
            self.job_timeouts.idle * (len(self.lithium.testcase) - 1)

        # iteration timeout is * n testcases, but add 10 seconds for overhead from the
        #   outer harness
        self.job_timeouts.iteration = \
            (self.job_timeouts.iteration + 10) * len(self.lithium.testcase)

    def on_success(self):
        # XXX: when py27 is dropped, this can be files_to_reduce.clear()
        while self.job_testcase.files_to_reduce:
            self.job_testcase.files_to_reduce.pop()
        lines = lithium.TestcaseLine()
        lines.readTestcase(self.job_testcase.entry)
        if len(lines) == 1:
            # we reduced this down to a single testcase, remove the harness
            testcase_rel = lines.parts[0].strip().decode("utf-8")
            assert testcase_rel.startswith("'/")
            assert testcase_rel.endswith("',")
            testcase_rel = testcase_rel[2:-2]  # trim chars asserted above
            testcase_path = testcase_rel.split('/')
            assert len(testcase_path) == 2
            self.job_testcase.root = os.path.join(self.job_testcase.root, testcase_path[0])
            self.job_testcase.entry = os.path.join(self.job_testcase.root, testcase_path[1])
            self.job_testcase.landing_page = self.job_testcase.entry
            self.job_testcase.files_to_reduce.append(self.job_testcase.entry)
            LOG.info("Reduced history to a single file: %s", testcase_path[1])
        else:
            # don't bother trying to reduce the harness further,
            #   leave harness out of files_to_reduce
            LOG.info("Reduced history down to %d testcases", len(self.lithium.testcase))


class ScanFilesToReduce(ReduceStage):
    name = "scan-files"
    ALTERS_JOB_TESTCASE = True

    # job attributes:
    # read: tcroot, testcase, files_to_reduce (len, iter)
    # modify: files_to_reduce (append, sort)
    # write: original_size

    def __init__(self, *args, **kwds):
        super(ScanFilesToReduce, self).__init__(*args, **kwds)
        # find all files for reduction
        for file_name in testcase_contents(self.job_testcase.root):
            file_name = os.path.join(self.job_testcase.root, file_name)
            if file_name == self.job_testcase.entry:
                continue
            with open(file_name, "rb") as tc_fp:
                for line in tc_fp:
                    if b"DDBEGIN" in line:
                        self.job_testcase.files_to_reduce.append(file_name)
                        break
        if len(self.job_testcase.files_to_reduce) > 1:
            # sort by descending size
            self.job_testcase.files_to_reduce.sort(key=lambda fn: os.stat(fn).st_size, reverse=True)
        self.job_testcase.original_size = self.job_testcase.total_size()

    def read_testcase(self, testcase_path):
        pass

    def should_skip(self):  # pylint: disable=no-self-argument
        # no real reduce strategy, just used to scan for files
        return True


class CSSBeautify(ReduceStage):
    name = "cssbeautify"
    strategy_type = lithium.CheckOnly
    testcase_type = lithium.TestcaseLine

    def __init__(self, *args, **kwds):
        super(CSSBeautify, self).__init__(*args, **kwds)
        self.testcase_path = None
        self.original_testcase = None
        self._ext = None
        self._force_skip = False

    def read_testcase(self, testcase_path):
        self.testcase_path = testcase_path
        self._ext = testcase_path.rsplit(".", 1)[-1]

        if self.should_skip():
            return

        # Beautify testcase
        opts = (
            ('end_with_newline', False),
            ('indent_size', 2),
            ('newline_between_rules', False),
            ('preserve_newlines', False),
        )
        with open(testcase_path) as testcase_fp:
            self.original_testcase = testcase_fp.read()
        if self._ext == "css":
            # DDBEGIN and DDEND are ignored here atm
            LOG.info("Attempting to cssbeautify %s", testcase_path)
            with open(testcase_path, "w") as testcase_fp:
                testcase_fp.write(cssbeautifier.beautify(self.original_testcase, opts))
        else:
            # handle html files
            begin = max(self.original_testcase.find("DDBEGIN"), 0)
            end = self.original_testcase.find("DDEND")
            if end == -1:
                end = len(self.original_testcase)
            re_tag = re.compile(r"(<style.*?>)(.*?)(</style>)", flags=re.DOTALL|re.IGNORECASE)
            if not re_tag.search(self.original_testcase, begin, end):
                LOG.debug("<style> tags not found in %r", testcase_path)
                self._force_skip = True
                self.original_testcase = None
                return
            pos = 0
            with open(testcase_path, "w") as testcase_fp:
                for match in re_tag.finditer(self.original_testcase, begin, end):
                    testcase_fp.write(self.original_testcase[pos:match.start(2)])
                    css = cssbeautifier.beautify(match.group(2), opts)
                    if css:
                        testcase_fp.write("\n")
                        testcase_fp.write(css)
                        testcase_fp.write("\n")
                    pos = match.end(2)
                testcase_fp.write(self.original_testcase[pos:])
            LOG.info("Ran cssbeautify on %s", testcase_path)

        self.lithium.strategy = self.strategy_type()  # pylint: disable=not-callable
        self.lithium.testcase = self.testcase_type()  # pylint: disable=not-callable
        self.lithium.testcase.readTestcase(testcase_path)

    def should_skip(self):
        if HAVE_CSSBEAUTIFIER and not self._force_skip:
            if self._ext in ("css", "htm", "html"):
                return False
        return True

    def on_failure(self):
        LOG.warning("CSSBeautification failed (reverting)")
        with open(self.testcase_path, "w") as testcase_fp:
            testcase_fp.write(self.original_testcase)


class JSBeautify(ReduceStage):
    name = "jsbeautify"
    strategy_type = lithium.CheckOnly
    testcase_type = lithium.TestcaseLine

    def __init__(self, *args, **kwds):
        super(JSBeautify, self).__init__(*args, **kwds)
        self.testcase_path = None
        self.original_testcase = None

    def read_testcase(self, testcase_path):
        self.testcase_path = testcase_path

        if self.should_skip():
            return

        LOG.info("Attempting to jsbeautify %s", testcase_path)

        self.lithium.strategy = self.strategy_type()  # pylint: disable=not-callable
        self.lithium.testcase = self.testcase_type()  # pylint: disable=not-callable

        # Beautify testcase
        with open(testcase_path) as testcase_fp:
            self.original_testcase = testcase_fp.read()

        beautified_testcase = jsbeautifier.beautify(self.original_testcase)
        # All try/catch pairs will be expanded on their own lines
        # Collapse these pairs when only a single instruction is contained
        #   within
        regex = r"(\s*try {)\n\s*(.*)\n\s*(}\s*catch.*)"
        beautified_testcase = re.sub(regex, r"\1 \2 \3", beautified_testcase)
        with open(testcase_path, 'w') as testcase_fp:
            testcase_fp.write(beautified_testcase)

        self.lithium.testcase.readTestcase(testcase_path)

    def should_skip(self):
        if HAVE_JSBEAUTIFIER and self.testcase_path.endswith(".js"):
            # jsbeautifier is only effective with JS files
            return False
        return True

    def on_failure(self):
        LOG.warning("JSBeautification failed (reverting)")
        with open(self.testcase_path, 'w') as testcase_fp:
            testcase_fp.write(self.original_testcase)


class CollapseEmptyBraces(ReduceStage):
    name = "collapsebraces"
    strategy_type = lithium.CollapseEmptyBraces
    testcase_type = lithium.TestcaseLine


class MinimizeChars(ReduceStage):
    name = "char"
    strategy_type = lithium.Minimize
    testcase_type = lithium.TestcaseChar


class MinimizeJSChars(ReduceStage):
    name = "jschar"
    strategy_type = lithium.Minimize
    testcase_type = lithium.TestcaseJsStr


def strategies_by_name():
    result = {}
    for cls in globals().values():
        if isinstance(cls, type) and issubclass(cls, ReduceStage) and cls is not ReduceStage:
            if cls.name in result:
                raise RuntimeError("Duplicate strategy name: %s" % (cls.name,))
            result[cls.name] = cls
    return result



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from argparse import ArgumentParser, HelpFormatter
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING
from os import listdir
from os.path import exists, isfile, isdir

from .adapters import names as adapter_names
from .target import available as available_targets

# ref: https://stackoverflow.com/questions/12268602/sort-argparse-help-alphabetically
class SortingHelpFormatter(HelpFormatter):
    @staticmethod
    def __sort_key(action):
        for opt in action.option_strings:
            if opt.startswith("--"):
                return [opt]
        return action.option_strings

    def add_usage(self, usage, actions, groups, prefix=None):
        actions = sorted(actions, key=self.__sort_key)
        super(SortingHelpFormatter, self).add_usage(usage, actions, groups, prefix)

    def add_arguments(self, actions):
        actions = sorted(actions, key=self.__sort_key)
        super(SortingHelpFormatter, self).add_arguments(actions)


class CommonArgs(object):
    IGNORABLE = ("log-limit", "memory", "timeout")
    IGNORE = ("log-limit", "timeout")

    def __init__(self):
        super(CommonArgs, self).__init__()
        # log levels for console logging
        self._level_map = {
            "CRIT": CRITICAL,
            "ERROR": ERROR,
            "WARN": WARNING,
            "INFO": INFO,
            "DEBUG": DEBUG}
        self._sanity_skip = set()

        if not hasattr(self, "parser"):
            self.parser = ArgumentParser(
                formatter_class=SortingHelpFormatter,
                conflict_handler='resolve')

        self.parser.add_argument(
            "binary",
            help="Firefox binary to run")
        self.parser.add_argument(
            "--log-level", default="INFO",
            help="Configure console logging. Options: %s (default: %%(default)s)" %
            ", ".join(k for k, v in sorted(self._level_map.items(), key=lambda x: x[1])))

        self.launcher_grp = self.parser.add_argument_group("Launcher Arguments")
        self.launcher_grp.add_argument(
            "-e", "--extension", action="append",
            help="Install an extension. Specify the path to the xpi or the directory"
                 " containing the unpacked extension. To install multiple extensions"
                 " specify multiple times")
        self.launcher_grp.add_argument(
            "--launch-timeout", type=int, default=300,
            help="Number of seconds to wait before LaunchError is raised (default: %(default)s)")
        self.launcher_grp.add_argument(
            "--log-limit", type=int, default=0,
            help="Browser log file size limit in MBs (default: 'no limit')")
        self.launcher_grp.add_argument(
            "-m", "--memory", type=int, default=0,
            help="Browser process memory limit in MBs (default: 'no limit')")
        self.launcher_grp.add_argument(
            "--platform", default="ffpuppet",
            help="Platforms available: %s (default: %%(default)s)" % ", ".join(available_targets()))
        self.launcher_grp.add_argument(
            "-p", "--prefs",
            help="prefs.js file to use")
        self.launcher_grp.add_argument(
            "--relaunch", type=int, default=1000,
            help="Number of iterations performed before relaunching the browser (default: %(default)s)")
        self.launcher_grp.add_argument(
            "-t", "--timeout", type=int, default=60,
            help="Iteration timeout in seconds (default: %(default)s)")
        self.launcher_grp.add_argument(
            "--valgrind", action="store_true",
            help="Use Valgrind (Linux only)")
        self.launcher_grp.add_argument(
            "--xvfb", action="store_true",
            help="Use Xvfb (Linux only)")

        self.reporter_grp = self.parser.add_argument_group("Reporter Arguments")
        self.reporter_grp.add_argument(
            "--fuzzmanager", action="store_true",
            help="Report results to FuzzManager")
        self.reporter_grp.add_argument(
            "--ignore", nargs="*", default=list(self.IGNORE),
            help="Space separated list of issue types to ignore. Valid options: %s"
                 " (default: %s)" % (" ".join(self.IGNORABLE), " ".join(self.IGNORE)))
        self.reporter_grp.add_argument(
            "--tool",
            help="Override tool name used when reporting issues to FuzzManager")

    def parse_args(self, argv=None):
        args = self.parser.parse_args(argv)
        self.sanity_check(args)
        return args

    def sanity_check(self, args):
        if hasattr(super(CommonArgs, self), 'sanity_check'):
            super(CommonArgs, self).sanity_check(args)

        if "binary" not in self._sanity_skip and not isfile(args.binary):
            self.parser.error("file not found: %r" % args.binary)

        # sanitize ignore list
        args.ignore = {arg.lower() for arg in args.ignore}
        for ignore in args.ignore:
            if ignore not in self.IGNORABLE:
                self.parser.error("Unrecognized ignore value: %s" % ignore)

        if "input" not in self._sanity_skip and args.input:
            if not exists(args.input):
                self.parser.error("%r does not exist" % args.input)
            elif isdir(args.input) and not listdir(args.input):
                self.parser.error("%r is empty" % args.input)

        # check log level
        log_level = self._level_map.get(args.log_level.upper(), None)
        if log_level is None:
            self.parser.error("Invalid log-level %r" % args.log_level)
        args.log_level = log_level

        if args.log_limit < 0:
            self.parser.error("--log-limit must be >= 0")
        args.log_limit *= 1048576

        if args.memory < 0:
            self.parser.error("-m/--memory must be >= 0")
        args.memory *= 1048576

        if args.relaunch < 1:
            self.parser.error("--relaunch must be >= 1")

        if args.extension:
            for ext in args.extension:
                if not exists(ext):
                    self.parser.error("%r does not exist" % ext)
                if not isdir(ext) or (isfile(ext) and ext.endswith(".xpi")):
                    self.parser.error("Extension must be a folder or .xpi")

        if args.platform.lower() not in set(available_targets()):
            self.parser.error("Unsupported platform %r" % args.platform)

        if args.prefs and not isfile(args.prefs):
            self.parser.error("-p/--prefs not found %r" % args.prefs)

        if "tool" not in self._sanity_skip:
            if args.tool is not None and not args.fuzzmanager:
                self.parser.error("--tool can only be given with --fuzzmanager")


class GrizzlyArgs(CommonArgs):
    def __init__(self):
        super(GrizzlyArgs, self).__init__()
        self._adapters = sorted(adapter_names())
        self._sanity_skip.add("tool")
        self.parser.add_argument(
            "adapter",
            help="Available adapters: %s" % ", ".join(self._adapters))
        self.parser.add_argument(
            "-i", "--input",
            help="Test case or directory containing test cases")
        self.parser.add_argument(
            "-v", "--verbose", action="store_true",
            help="Output console updates every iteration. By default the number"
                 " of iterations between console updates doubles each update."
                 " Updates are always printed when a result is detected or the"
                 " target is relaunched.")

        self.launcher_grp.add_argument(
            "--coverage", action="store_true",
            help="Enable coverage collection")
        self.launcher_grp.add_argument(
            "--rr", action="store_true",
            help="Use RR (Linux only)")

        self.reporter_grp.add_argument(
            "-c", "--cache", type=int, default=0,
            help="Maximum number of additional test cases to include in report (default: %(default)s)")
        self.reporter_grp.add_argument(
            "--s3-fuzzmanager", action="store_true",
            help="Report large attachments (if any) to S3 and then the crash & S3 link to FuzzManager")

        self.parser.epilog = "For addition help check out the wiki:" \
            " https://github.com/MozillaSecurity/grizzly/wiki"

    def sanity_check(self, args):
        super(GrizzlyArgs, self).sanity_check(args)

        if args.adapter.lower() not in self._adapters:
            msg = ["Adapter %r does not exist." % args.adapter.lower()]
            if self._adapters:
                msg.append("Available adapters: %s" % ", ".join(self._adapters))
            else:
                msg.append("No adapters available.")
            self.parser.error(" ".join(msg))

        if args.fuzzmanager and args.s3_fuzzmanager:
            self.parser.error("--fuzzmanager and --s3-fuzzmanager are mutually exclusive")

        if args.tool is not None and not (args.fuzzmanager or args.s3_fuzzmanager):
            self.parser.error("--tool can only be given with --fuzzmanager/--s3-fuzzmanager")



# coding: utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
import random
import struct
import pytest

from .loki import Loki


def test_loki_01(tmp_path):
    """test a missing file"""
    fuzzer = Loki(aggression=0.1)
    fuzzer.fuzz_file("nofile.test", 1, out_dir=str(tmp_path))
    assert not list(tmp_path.iterdir())


def test_loki_02(tmp_path):
    """test an empty file"""
    tmp_fn = tmp_path / "input"
    tmp_fn.touch()

    out_path = tmp_path / "out"
    out_path.mkdir()

    fuzzer = Loki(aggression=0.1)
    fuzzer.fuzz_file(str(tmp_fn), 1, out_dir=str(out_path))
    assert not list(out_path.iterdir())


def test_loki_03(tmp_path):
    """test a single byte file"""
    in_data = b"A"
    tmp_fn = tmp_path / "input"
    tmp_fn.write_bytes(in_data)

    out_path = tmp_path / "out"
    out_path.mkdir()

    fuzzer = Loki(aggression=0.1)
    for _ in range(100):
        fuzzer.fuzz_file(str(tmp_fn), 1, out_dir=str(out_path))
        out_files = list(out_path.iterdir())
        assert len(out_files) == 1
        out_data = out_files[0].read_bytes()
        assert len(out_data) == 1
        if out_data != in_data:
            break
    assert out_data != in_data


def test_loki_04(tmp_path):
    """test a two byte file"""
    in_data = b"AB"
    tmp_fn = tmp_path / "input"
    tmp_fn.write_bytes(in_data)

    out_path = tmp_path / "out"
    out_path.mkdir()

    fuzzer = Loki(aggression=0.1)
    for _ in range(100):
        fuzzer.fuzz_file(str(tmp_fn), 1, out_dir=str(out_path))
        out_files = list(out_path.iterdir())
        assert len(out_files) == 1
        out_data = out_files[0].read_bytes()
        assert len(out_data) == 2
        if out_data != in_data:
            break
    assert out_data != in_data


def test_loki_05(tmp_path):
    """test a multi byte file"""
    in_size = 100
    in_byte = b"A"
    in_data = in_byte * in_size
    fuzz_found = False
    tmp_fn = tmp_path / "input"
    tmp_fn.write_bytes(in_data)

    out_path = tmp_path / "out"
    out_path.mkdir()

    fuzzer = Loki(aggression=0.01)
    for _ in range(100):
        fuzzer.fuzz_file(str(tmp_fn), 1, out_dir=str(out_path))
        out_files = list(out_path.iterdir())
        assert len(out_files) == 1
        with out_files[0].open("rb") as out_fp:
            out_fp.seek(0, os.SEEK_END)
            assert out_fp.tell() == in_size
            out_fp.seek(0)
            for out_byte in out_fp:
                if out_byte != in_byte:
                    fuzz_found = True
                    break
        if fuzz_found:
            break
    assert fuzz_found


def test_loki_06():
    """test fuzz_data()"""
    in_data = b"This is test DATA!"
    in_size = len(in_data)

    fuzz_found = False
    fuzzer = Loki(aggression=0.1)
    for _ in range(100):
        out_data = fuzzer.fuzz_data(in_data)
        assert len(out_data) == in_size
        if in_data not in out_data:
            fuzz_found = True
            break
    assert fuzz_found


def test_loki_07():
    """test invalid data sizes"""
    with pytest.raises(RuntimeError, match=r"Unsupported data size:"):
        Loki._fuzz_data(b"")  # pylint: disable=protected-access

    with pytest.raises(RuntimeError, match=r"Unsupported data size:"):
        Loki._fuzz_data(b"123")  # pylint: disable=protected-access

    with pytest.raises(RuntimeError, match=r"Unsupported data size:"):
        Loki._fuzz_data(b"12345")  # pylint: disable=protected-access


def test_loki_08():
    """test endian support"""
    Loki._fuzz_data(b"1", ">")  # pylint: disable=protected-access
    Loki._fuzz_data(b"1", "<")  # pylint: disable=protected-access
    with pytest.raises(RuntimeError, match=r"Unsupported byte order"):
        Loki._fuzz_data(b"1", "BAD")  # pylint: disable=protected-access


def test_loki_stress_01():
    """test with single byte"""
    for _ in range(1000):
        in_data = struct.pack("B", random.getrandbits(8))
        assert len(Loki._fuzz_data(in_data)) == 1  # pylint: disable=protected-access


def test_loki_stress_02():
    """test with two bytes"""
    in_data = b"\xff\xff"
    for _ in range(1000):
        assert len(Loki._fuzz_data(in_data)) == 2  # pylint: disable=protected-access


def test_loki_stress_03():
    """test with four bytes"""
    in_data = b"TEST"
    for _ in range(1000):
        assert len(Loki._fuzz_data(in_data)) == 4  # pylint: disable=protected-access



#!/usr/bin/env python
# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""Grizzly setup"""
from os.path import dirname, join as pathjoin
from itertools import chain
from setuptools import setup


EXTRAS = {
    'reduce': ['lithium-reducer', 'jsbeautifier'],
    's3': ['boto3'],
}
EXTRAS['all'] = list(set(chain.from_iterable(EXTRAS.values())))
EXTRAS['test'] = ['pytest>=3.9', 'pytest-cov', 'pytest-mock']


if __name__ == '__main__':
    with open(pathjoin(dirname(__file__), 'README.md'), 'r') as infp:
        README = infp.read()
    setup(
        classifiers=[
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Testing',
            'Topic :: Security',
            'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8'
        ],
        description='A framework for developing and running browser fuzzers',
        entry_points={
            'console_scripts': [
                'grizzly = grizzly.main:console_main',
                'grizzly.status = grizzly.common.status_reporter:main',
            ],
            'grizzly_targets': [
                'ffpuppet = grizzly.target.puppet_target:PuppetTarget',
            ],
        },
        extras_require=EXTRAS,
        install_requires=[
            'fasteners',
            'ffpuppet',
            'FuzzManager',
            'prefpicker',
            'psutil'
        ],
        keywords='firefox framework fuzz fuzzing test testing',
        license='MPL 2.0',
        long_description=README,
        long_description_content_type='text/markdown',
        maintainer='Mozilla Fuzzing Team',
        maintainer_email='fuzzing@mozilla.com',
        name='grizzly-framework',
        packages=[
            'grizzly',
            'grizzly.adapters',
            'grizzly.adapters.NoOpAdapter',
            'grizzly.common',
            'grizzly.reduce',
            'grizzly.replay',
            'grizzly.target',
            'loki',
            'sapphire',
        ],
        package_data={"grizzly.common": ["harness.html"]},
        url='https://github.com/MozillaSecurity/grizzly',
        version='0.9.6')



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

SERVED_ALL = 0      # all expected requests for required files have been received
SERVED_NONE = 1     # no requests for required files have been received
SERVED_REQUEST = 2  # some requests for required files have been received
SERVED_TIMEOUT = 3  # timeout occurred
